# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tempfile

import numpy as np
import pandas as pd
import pytest
import torch
from packaging import version

from ludwig.api import LudwigModel
from ludwig.backend import create_ray_backend, initialize_backend, LOCAL_BACKEND
from ludwig.constants import (
    BALANCE_PERCENTAGE_TOLERANCE,
    BFILL,
    COLUMN,
    DEFAULT_BATCH_SIZE,
    NAME,
    PREPROCESSING,
    TRAINER,
)
from ludwig.data.preprocessing import balance_data
from ludwig.utils.data_utils import read_parquet
from tests.integration_tests.utils import (
    audio_feature,
    augment_dataset_with_none,
    bag_feature,
    binary_feature,
    category_feature,
    create_data_set_to_use,
    date_feature,
    generate_data,
    h3_feature,
    image_feature,
    number_feature,
    RAY_BACKEND_CONFIG,
    sequence_feature,
    set_feature,
    text_feature,
    timeseries_feature,
    train_with_backend,
    vector_feature,
)

try:
    import modin
    import ray

    from ludwig.backend.ray import get_trainer_kwargs, RayBackend
    from ludwig.data.dataframe.dask import DaskEngine

    @ray.remote(num_cpus=1, num_gpus=1)
    def train_gpu(config, dataset, output_directory):
        model = LudwigModel(config, backend="local")
        _, _, output_dir = model.train(dataset, output_directory=output_directory)
        return os.path.join(output_dir, "model")

    @ray.remote(num_cpus=1, num_gpus=0)
    def predict_cpu(model_dir, dataset):
        model = LudwigModel.load(model_dir, backend="local")
        model.predict(dataset)

    # Ray nightly version is always set to 3.0.0.dev0
    _ray_nightly = version.parse(ray.__version__) >= version.parse("3.0.0.dev0")
    _modin_ray_incompatible = version.parse(modin.__version__) <= version.parse("0.15.2") and version.parse(
        ray.__version__
    ) >= version.parse("1.13.0")

except ImportError:
    modin = None
    ray = None

    _ray_nightly = False
    _modin_ray_incompatible = False


def run_api_experiment(
    config, dataset, backend_config, predict=False, skip_save_processed_input=True, skip_save_predictions=True
):
    # Sanity check that we get 4 slots over 1 host
    kwargs = get_trainer_kwargs()
    if torch.cuda.device_count() > 0:
        assert kwargs.get("num_workers") == torch.cuda.device_count(), kwargs
        assert kwargs.get("use_gpu"), kwargs
    else:
        assert kwargs.get("num_workers") == 1, kwargs
        assert not kwargs.get("use_gpu"), kwargs

    # Train on Parquet
    model = train_with_backend(
        backend_config,
        config,
        dataset=dataset,
        evaluate=True,
        predict=predict,
        skip_save_processed_input=skip_save_processed_input,
        skip_save_predictions=skip_save_predictions,
    )

    assert isinstance(model.backend, RayBackend)
    if isinstance(model.backend.df_engine, DaskEngine):
        assert model.backend.df_engine.parallelism == backend_config["processor"]["parallelism"]

    return model


def run_split_api_experiment(config, data_parquet, backend_config):
    train_fname, val_fname, test_fname = split(data_parquet)

    # Train
    train_with_backend(backend_config, config, training_set=train_fname, evaluate=False, predict=True)

    # Train + Validation
    train_with_backend(
        backend_config, config, training_set=train_fname, validation_set=val_fname, evaluate=False, predict=False
    )

    # Train + Validation + Test
    train_with_backend(
        backend_config,
        config,
        training_set=train_fname,
        validation_set=val_fname,
        test_set=test_fname,
        evaluate=False,
        predict=False,
    )


def split(data_parquet):
    data_df = read_parquet(data_parquet, LOCAL_BACKEND.df_engine.df_lib)
    train_df = data_df.sample(frac=0.8)
    test_df = data_df.drop(train_df.index).sample(frac=0.5)
    validation_df = data_df.drop(train_df.index).drop(test_df.index)

    basename, ext = os.path.splitext(data_parquet)
    train_fname = basename + ".train" + ext
    val_fname = basename + ".validation" + ext
    test_fname = basename + ".test" + ext

    train_df.to_parquet(train_fname)
    validation_df.to_parquet(val_fname)
    test_df.to_parquet(test_fname)
    return train_fname, val_fname, test_fname


def run_test_with_features(
    input_features,
    output_features,
    num_examples=100,
    run_fn=run_api_experiment,
    expect_error=False,
    df_engine=None,
    dataset_type="parquet",
    predict=False,
    skip_save_processed_input=True,
    skip_save_predictions=True,
    nan_percent=0.0,
    preprocessing=None,
    first_row_none=False,
    last_row_none=False,
    nan_cols=[],
):
    preprocessing = preprocessing or {}
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": 2, "batch_size": 8},
    }
    if preprocessing:
        config[PREPROCESSING] = preprocessing

    backend_config = {**RAY_BACKEND_CONFIG}
    if df_engine:
        backend_config["processor"]["type"] = df_engine

    with tempfile.TemporaryDirectory() as tmpdir:
        csv_filename = os.path.join(tmpdir, "dataset.csv")
        dataset_csv = generate_data(input_features, output_features, csv_filename, num_examples=num_examples)
        dataset = create_data_set_to_use(dataset_type, dataset_csv, nan_percent=nan_percent)
        dataset = augment_dataset_with_none(dataset, first_row_none, last_row_none, nan_cols)

        if expect_error:
            with pytest.raises(ValueError):
                run_fn(
                    config,
                    dataset=dataset,
                    backend_config=backend_config,
                    predict=predict,
                    skip_save_processed_input=skip_save_processed_input,
                    skip_save_predictions=skip_save_predictions,
                )
        else:
            run_fn(
                config,
                dataset=dataset,
                backend_config=backend_config,
                predict=predict,
                skip_save_processed_input=skip_save_processed_input,
                skip_save_predictions=skip_save_predictions,
            )


@pytest.mark.parametrize("df_engine", ["pandas", "dask"])
@pytest.mark.distributed
def test_ray_read_binary_files(tmpdir, df_engine, ray_cluster_2cpu):
    preprocessing_params = {
        "audio_file_length_limit_in_s": 3.0,
        "missing_value_strategy": BFILL,
        "in_memory": True,
        "padding_value": 0,
        "norm": "per_file",
        "audio_feature": {
            "type": "fbank",
            "window_length_in_s": 0.04,
            "window_shift_in_s": 0.02,
            "num_filter_bands": 80,
        },
    }
    audio_dest_folder = os.path.join(tmpdir, "generated_audio")
    audio_params = audio_feature(folder=audio_dest_folder, preprocessing=preprocessing_params)

    dataset_path = os.path.join(tmpdir, "dataset.csv")
    dataset_path = generate_data([audio_params], [], dataset_path, num_examples=10)
    dataset_path = create_data_set_to_use("csv", dataset_path, nan_percent=0.1)

    backend_config = {**RAY_BACKEND_CONFIG}
    backend_config["processor"]["type"] = df_engine
    backend = initialize_backend(backend_config)
    df = backend.df_engine.df_lib.read_csv(dataset_path)
    series = df[audio_params[COLUMN]]
    proc_col = backend.read_binary_files(series)
    proc_col = backend.df_engine.compute(proc_col)

    backend = initialize_backend(LOCAL_BACKEND)
    df = backend.df_engine.df_lib.read_csv(dataset_path)
    series = df[audio_params[COLUMN]]
    proc_col_expected = backend.read_binary_files(series)

    assert proc_col.equals(proc_col_expected)


# TODO(geoffrey): Add dataset_type="csv" back to parameters if we can prevent CI timeouts.
@pytest.mark.parametrize("dataset_type", ["parquet"])
@pytest.mark.distributed
def test_ray_save_inputs_with_nans(tmpdir, dataset_type, ray_cluster_2cpu):
    image_dest_folder = os.path.join(tmpdir, "generated_images")
    audio_dest_folder = os.path.join(tmpdir, "generated_audio")
    input_features = [
        image_feature(
            folder=image_dest_folder,
            preprocessing={"in_memory": True, "height": 12, "width": 12, "num_channels": 3, "num_processes": 5},
            encoder={"output_size": 16, "num_filters": 8},
        ),
        audio_feature(
            folder=audio_dest_folder,
            preprocessing={
                "audio_file_length_limit_in_s": 3.0,
                "missing_value_strategy": BFILL,
                "in_memory": True,
                "padding_value": 0,
                "norm": "per_file",
                "type": "fbank",
                "window_length_in_s": 0.04,
                "window_shift_in_s": 0.02,
                "num_filter_bands": 80,
            },
        ),
        sequence_feature(encoder={"reduce_output": "sum"}),
        category_feature(encoder={"vocab_size": 2}, reduce_input="sum"),
        number_feature(normalization="zscore"),
        set_feature(),
        binary_feature(),
        bag_feature(),
        text_feature(),
        timeseries_feature(),
        date_feature(),
        # TODO: NaN handling not supported. See `test_ray_save_inputs_and_outputs_without_nans` below.
        # vector_feature(),  # NaNs are not supported by the feature
        # TODO: feature type not yet supported
        # h3_feature(),  # ValueError casting large int strings (e.g. '5.864041857092157e+17') to int: MLI-72
    ]
    output_features = [
        category_feature(decoder={"vocab_size": 5}),  # Regression test for #1991 requires multi-class predictions.
    ]
    run_test_with_features(
        input_features,
        output_features,
        df_engine="dask",
        dataset_type=dataset_type,
        skip_save_processed_input=False,
        nan_percent=0.1,
    )


@pytest.mark.parametrize("dataset_type", ["csv", "parquet"])
@pytest.mark.distributed
def test_ray_save_inputs_without_nans(dataset_type, ray_cluster_2cpu):
    input_features = [
        vector_feature(),
    ]
    output_features = [
        binary_feature(),
    ]
    run_test_with_features(
        input_features,
        output_features,
        df_engine="dask",
        dataset_type=dataset_type,
        skip_save_processed_input=False,
    )


@pytest.mark.parametrize("dataset_type", ["csv", "parquet"])
@pytest.mark.distributed
def test_ray_save_outputs(dataset_type, ray_cluster_2cpu):
    input_features = [
        binary_feature(),
    ]
    output_features = [
        binary_feature(),
        number_feature(),
        vector_feature(),
        # TODO: feature type not yet supported
        # set_feature(decoder={"vocab_size": 3}),  # Probabilities of set_feature are ragged tensors: MLI-71
        # sequence_feature(decoder={"vocab_size": 3}),  # Error having to do with a missing key: MLI-70
        # text_feature(decoder={"vocab_size": 3}),      # Error having to do with a missing key: MLI-70
    ]
    # NOTE: This test runs without NaNs because having multiple output features with DROP_ROWS strategy leads to
    # flakiness in the test having to do with uneven allocation of samples between Ray workers.
    run_test_with_features(
        input_features,
        output_features,
        df_engine="dask",
        dataset_type=dataset_type,
        predict=True,
        skip_save_predictions=False,
    )


@pytest.mark.distributed
@pytest.mark.parametrize(
    "df_engine",
    [
        "dask",
        pytest.param(
            "modin",
            marks=pytest.mark.skipif(_modin_ray_incompatible, reason="modin<=0.15.2 does not support ray>=1.13.0"),
        ),
    ],
)
def test_ray_tabular(df_engine, ray_cluster_2cpu):
    input_features = [
        sequence_feature(encoder={"reduce_output": "sum"}),
        category_feature(encoder={"vocab_size": 2}, reduce_input="sum"),
        number_feature(normalization="zscore"),
        set_feature(),
        binary_feature(),
        bag_feature(),
        vector_feature(),
        h3_feature(),
        date_feature(),
    ]
    output_features = [
        binary_feature(bool2str=["No", "Yes"]),
        binary_feature(),
        number_feature(normalization="zscore"),
    ]
    run_test_with_features(
        input_features,
        output_features,
        df_engine=df_engine,
    )


@pytest.mark.skip(reason="TODO torch")
@pytest.mark.distributed
def test_ray_text(ray_cluster_2cpu):
    input_features = [
        text_feature(),
    ]
    output_features = [
        text_feature(reduce_input=None, decoder={"type": "tagger"}),
    ]
    run_test_with_features(input_features, output_features)


@pytest.mark.skip(reason="TODO torch")
@pytest.mark.distributed
def test_ray_sequence(ray_cluster_2cpu):
    input_features = [
        sequence_feature(encoder={"max_len": 10, "type": "rnn", "cell_type": "lstm", "reduce_output": None})
    ]
    output_features = [
        sequence_feature(decoder={"max_len": 10, "type": "tagger", "attention": False}, reduce_input=None)
    ]
    run_test_with_features(input_features, output_features)


@pytest.mark.parametrize("dataset_type", ["csv", "parquet"])
@pytest.mark.distributed
def test_ray_audio(tmpdir, dataset_type, ray_cluster_2cpu):
    preprocessing_params = {
        "audio_file_length_limit_in_s": 3.0,
        "missing_value_strategy": BFILL,
        "in_memory": True,
        "padding_value": 0,
        "norm": "per_file",
        "type": "fbank",
        "window_length_in_s": 0.04,
        "window_shift_in_s": 0.02,
        "num_filter_bands": 80,
    }
    audio_dest_folder = os.path.join(tmpdir, "generated_audio")
    input_features = [audio_feature(folder=audio_dest_folder, preprocessing=preprocessing_params)]
    output_features = [binary_feature()]
    run_test_with_features(
        input_features,
        output_features,
        dataset_type=dataset_type,
        nan_percent=0.1,
    )


@pytest.mark.parametrize("dataset_type", ["csv", "parquet", "pandas+numpy_images"])
@pytest.mark.distributed
def test_ray_image(tmpdir, dataset_type, ray_cluster_2cpu):
    image_dest_folder = os.path.join(tmpdir, "generated_images")
    input_features = [
        image_feature(
            folder=image_dest_folder,
            preprocessing={"in_memory": True, "height": 12, "width": 12, "num_channels": 3, "num_processes": 5},
            encoder={"output_size": 16, "num_filters": 8},
        ),
    ]
    output_features = [binary_feature()]
    run_test_with_features(
        input_features,
        output_features,
        df_engine="dask",
        dataset_type=dataset_type,
        skip_save_processed_input=False,
        nan_percent=0.1,
    )


@pytest.mark.parametrize(
    "settings",
    [(True, False, "ffill"), (False, True, "bfill"), (True, True, "bfill"), (True, True, "ffill")],
    ids=["first_row_none", "last_row_none", "first_and_last_row_none_bfill", "first_and_last_row_none_ffill"],
)
@pytest.mark.distributed
def test_ray_image_with_fill_strategy_edge_cases(tmpdir, settings, ray_cluster_2cpu):
    first_row_none, last_row_none, missing_value_strategy = settings
    image_dest_folder = os.path.join(tmpdir, "generated_images")
    input_features = [
        image_feature(
            folder=image_dest_folder,
            preprocessing={
                "in_memory": True,
                "height": 12,
                "width": 12,
                "num_channels": 3,
                "num_processes": 5,
                "missing_value_strategy": missing_value_strategy,
            },
            encoder={"output_size": 16, "num_filters": 8},
        ),
    ]
    output_features = [binary_feature()]
    run_test_with_features(
        input_features,
        output_features,
        df_engine="dask",
        dataset_type="pandas+numpy_images",
        skip_save_processed_input=False,
        first_row_none=first_row_none,
        last_row_none=last_row_none,
        nan_cols=[input_features[0][NAME]],
    )


# TODO(geoffrey): Fold modin tests into test_ray_image as @pytest.mark.parametrized once tests are optimized
@pytest.mark.distributed
@pytest.mark.skipif(_modin_ray_incompatible, reason="modin<=0.15.2 does not support ray>=1.13.0")
def test_ray_image_modin(tmpdir, ray_cluster_2cpu):
    image_dest_folder = os.path.join(tmpdir, "generated_images")
    input_features = [
        image_feature(
            folder=image_dest_folder,
            encoder={"type": "resnet", "output_size": 16, "num_filters": 8},
            preprocessing={"in_memory": True, "height": 12, "width": 12, "num_channels": 3, "num_processes": 5},
        ),
    ]
    output_features = [binary_feature()]
    run_test_with_features(
        input_features,
        output_features,
        df_engine="modin",
        dataset_type="csv",
        nan_percent=0.1,
    )


@pytest.mark.distributed
def test_ray_image_multiple_features(tmpdir, ray_cluster_2cpu):
    input_features = [
        image_feature(
            folder=os.path.join(tmpdir, "generated_images_1"),
            preprocessing={"in_memory": True, "height": 12, "width": 12, "num_channels": 3, "num_processes": 5},
            encoder={"output_size": 16, "num_filters": 8},
        ),
        image_feature(
            folder=os.path.join(tmpdir, "generated_images_2"),
            preprocessing={"in_memory": True, "height": 12, "width": 12, "num_channels": 3, "num_processes": 5},
            encoder={"output_size": 16, "num_filters": 8},
        ),
    ]
    output_features = [binary_feature()]
    run_test_with_features(
        input_features,
        output_features,
        df_engine="dask",
        dataset_type="csv",
        nan_percent=0.1,
    )


@pytest.mark.skip(reason="flaky: ray is running out of resources")
@pytest.mark.distributed
def test_ray_split(ray_cluster_2cpu):
    input_features = [
        number_feature(normalization="zscore"),
        set_feature(),
        binary_feature(),
    ]
    output_features = [category_feature(decoder={"vocab_size": 2}, reduce_input="sum")]
    run_test_with_features(
        input_features,
        output_features,
        run_fn=run_split_api_experiment,
    )


@pytest.mark.distributed
def test_ray_timeseries(ray_cluster_2cpu):
    input_features = [timeseries_feature()]
    output_features = [number_feature()]
    run_test_with_features(input_features, output_features)


@pytest.mark.distributed
def test_ray_lazy_load_audio_error(tmpdir, ray_cluster_2cpu):
    audio_dest_folder = os.path.join(tmpdir, "generated_audio")
    input_features = [
        audio_feature(
            folder=audio_dest_folder,
            preprocessing={
                "in_memory": False,
            },
        )
    ]
    output_features = [binary_feature()]
    run_test_with_features(input_features, output_features, expect_error=True)


@pytest.mark.distributed
def test_ray_lazy_load_image_error(tmpdir, ray_cluster_2cpu):
    image_dest_folder = os.path.join(tmpdir, "generated_images")
    input_features = [
        image_feature(
            folder=image_dest_folder,
            encoder={"type": "resnet", "output_size": 16, "num_filters": 8},
            preprocessing={"in_memory": False, "height": 12, "width": 12, "num_channels": 3, "num_processes": 5},
        ),
    ]
    output_features = [binary_feature()]
    run_test_with_features(input_features, output_features, expect_error=True)


# TODO(travis): move this to separate gpu module so we only have one ray cluster running at a time
# @pytest.mark.skipif(torch.cuda.device_count() == 0, reason="test requires at least 1 gpu")
# @pytest.mark.skipif(not torch.cuda.is_available(), reason="test requires gpu support")
# @pytest.mark.distributed
# def test_train_gpu_load_cpu(ray_cluster_2cpu):
#     input_features = [
#         category_feature(encoder={"vocab_size": 2}, reduce_input="sum"),
#         number_feature(normalization="zscore"),
#     ]
#     output_features = [
#         binary_feature(),
#     ]
#     run_test_with_features(input_features, output_features, run_fn=_run_train_gpu_load_cpu, num_gpus=1)


@pytest.mark.distributed
@pytest.mark.parametrize(
    "method, balance",
    [
        ("oversample_minority", 0.25),
        ("oversample_minority", 0.5),
        ("oversample_minority", 0.75),
        ("undersample_majority", 0.25),
        ("undersample_majority", 0.5),
        ("undersample_majority", 0.75),
    ],
)
def test_balance_ray(method, balance, ray_cluster_2cpu):
    config = {
        "input_features": [
            {"name": "Index", "proc_column": "Index", "type": "number"},
            {"name": "random_1", "proc_column": "random_1", "type": "number"},
            {"name": "random_2", "proc_column": "random_2", "type": "number"},
        ],
        "output_features": [{"name": "Label", "proc_column": "Label", "type": "binary"}],
        "preprocessing": {"oversample_minority": None, "undersample_majority": None},
    }
    input_df = pd.DataFrame(
        {
            "Index": np.arange(0, 200, 1),
            "random_1": np.random.randint(0, 50, 200),
            "random_2": np.random.choice(["Type A", "Type B", "Type C", "Type D"], 200),
            "Label": np.concatenate((np.zeros(180), np.ones(20))),
            "split": np.zeros(200),
        }
    )
    config["preprocessing"][method] = balance
    target = config["output_features"][0][NAME]

    backend = create_ray_backend()
    input_df = backend.df_engine.from_pandas(input_df)
    test_df = balance_data(input_df, config["output_features"], config["preprocessing"], backend)

    majority_class = test_df[target].value_counts().compute()[test_df[target].value_counts().compute().idxmax()]
    minority_class = test_df[target].value_counts().compute()[test_df[target].value_counts().compute().idxmin()]
    new_class_balance = round(minority_class / majority_class, 2)

    assert abs(balance - new_class_balance) < BALANCE_PERCENTAGE_TOLERANCE


def _run_train_gpu_load_cpu(config, data_parquet):
    with tempfile.TemporaryDirectory() as output_dir:
        model_dir = ray.get(train_gpu.remote(config, data_parquet, output_dir))
        ray.get(predict_cpu.remote(model_dir, data_parquet))


# TODO(geoffrey): add a GPU test for batch size tuning
@pytest.mark.distributed
def test_tune_batch_size_lr_cpu(tmpdir, ray_cluster_2cpu):
    config = {
        "input_features": [
            number_feature(normalization="zscore"),
            set_feature(),
            binary_feature(),
        ],
        "output_features": [category_feature(decoder={"vocab_size": 2}, reduce_input="sum")],
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": 2, "batch_size": "auto", "learning_rate": "auto"},
    }

    backend_config = {**RAY_BACKEND_CONFIG}

    csv_filename = os.path.join(tmpdir, "dataset.csv")
    dataset_csv = generate_data(config["input_features"], config["output_features"], csv_filename, num_examples=200)
    dataset_parquet = create_data_set_to_use("parquet", dataset_csv)
    model = run_api_experiment(config, dataset=dataset_parquet, backend_config=backend_config)
    assert (
        model.config[TRAINER]["batch_size"] == DEFAULT_BATCH_SIZE
    )  # On CPU, batch size tuning is disabled, so assert it is equal to default
    assert model.config[TRAINER]["learning_rate"] != "auto"


@pytest.mark.distributed
def test_ray_progress_bar(ray_cluster_2cpu):
    # This is a simple test that is just meant to make sure that the progress bar isn't breaking
    input_features = [
        sequence_feature(encoder={"reduce_output": "sum"}),
    ]
    output_features = [
        binary_feature(bool2str=["No", "Yes"]),
    ]
    run_test_with_features(
        input_features,
        output_features,
        df_engine="dask",
    )


@pytest.mark.parametrize("calibration", [True, False])
@pytest.mark.distributed
def test_ray_calibration(calibration, ray_cluster_2cpu):
    input_features = [
        number_feature(normalization="zscore"),
        set_feature(),
        binary_feature(),
    ]
    output_features = [
        binary_feature(calibration=calibration),
        category_feature(decoder={"vocab_size": 3}, calibration=calibration),
    ]
    run_test_with_features(input_features, output_features)


@pytest.mark.distributed
def test_ray_distributed_predict(tmpdir, ray_cluster_2cpu):
    preprocessing_params = {
        "audio_file_length_limit_in_s": 3.0,
        "missing_value_strategy": BFILL,
        "in_memory": True,
        "padding_value": 0,
        "norm": "per_file",
        "type": "fbank",
        "window_length_in_s": 0.04,
        "window_shift_in_s": 0.02,
        "num_filter_bands": 80,
    }
    audio_dest_folder = os.path.join(tmpdir, "generated_audio")
    input_features = [audio_feature(folder=audio_dest_folder, preprocessing=preprocessing_params)]
    output_features = [binary_feature()]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"epochs": 2, "batch_size": 8},
    }

    with tempfile.TemporaryDirectory() as tmpdir:
        backend_config = {**RAY_BACKEND_CONFIG}
        csv_filename = os.path.join(tmpdir, "dataset.csv")
        dataset_csv = generate_data(input_features, output_features, csv_filename, num_examples=100)
        dataset = create_data_set_to_use("csv", dataset_csv, nan_percent=0.0)
        model = LudwigModel(config, backend=backend_config)
        output_dir = None

        _, _, output_dir = model.train(
            dataset=dataset,
            training_set=dataset,
            skip_save_processed_input=True,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
            skip_save_log=True,
        )

        preds, _ = model.predict(dataset=dataset)

        # compute the predictions
        preds = preds.compute()
        assert preds.iloc[1].name != preds.iloc[42].name


@pytest.mark.distributed
def test_ray_preprocessing_placement_group(tmpdir, ray_cluster_2cpu):
    preprocessing_params = {
        "audio_file_length_limit_in_s": 3.0,
        "missing_value_strategy": BFILL,
        "in_memory": True,
        "padding_value": 0,
        "norm": "per_file",
        "type": "fbank",
        "window_length_in_s": 0.04,
        "window_shift_in_s": 0.02,
        "num_filter_bands": 80,
    }
    audio_dest_folder = os.path.join(tmpdir, "generated_audio")
    input_features = [audio_feature(folder=audio_dest_folder, preprocessing=preprocessing_params)]
    output_features = [binary_feature()]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"epochs": 2, "batch_size": 8},
    }

    with tempfile.TemporaryDirectory() as tmpdir:
        backend_config = {**RAY_BACKEND_CONFIG}
        backend_config["preprocessor_kwargs"] = {"num_cpu": 1}
        csv_filename = os.path.join(tmpdir, "dataset.csv")
        dataset_csv = generate_data(input_features, output_features, csv_filename, num_examples=100)
        dataset = create_data_set_to_use("csv", dataset_csv, nan_percent=0.0)
        model = LudwigModel(config, backend=backend_config)
        _, _, output_dir = model.train(
            dataset=dataset,
            training_set=dataset,
            skip_save_processed_input=True,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
            skip_save_log=True,
        )
        preds, _ = model.predict(dataset=dataset)



from dataclasses import dataclass, field
from typing import List

import numpy as np
import numpy.typing as npt


@dataclass
class LabelExplanation:
    """Stores the feature attributions for a single label in the target feature's vocab."""

    # The attribution for each input feature.
    feature_attributions: npt.NDArray[np.float64]


@dataclass
class Explanation:
    """Stores the explanations for a single row of input data.

    Contains the feature attributions for each label in the target feature's vocab.
    """

    target: str

    # The explanations for each label in the vocab of the target feature.
    label_explanations: List[LabelExplanation] = field(default_factory=list)

    def add(self, feature_attributions: npt.NDArray[np.float64]):
        """Add the feature attributions for a single label."""
        if len(self.label_explanations) > 0:
            # Check that the feature attributions are the same shape as existing explanations.
            assert self.label_explanations[0].feature_attributions.shape == feature_attributions.shape, (
                f"Expected feature attributions of shape {self.label_explanations[0].feature_attributions.shape}, "
                f"got {feature_attributions.shape}"
            )
        self.label_explanations.append(LabelExplanation(feature_attributions))

    def to_array(self) -> npt.NDArray[np.float64]:
        """Convert the explanation to a 2D array of shape (num_labels, num_features)."""
        return np.array([le.feature_attributions for le in self.label_explanations])



from typing import Dict, List, Type, Union

from ludwig.encoders.base import Encoder
from ludwig.utils.registry import Registry

encoder_registry = Registry()

sequence_encoder_registry = Registry()


def register_sequence_encoder(name: str):
    def wrap(cls):
        sequence_encoder_registry[name] = cls
        return cls

    return wrap


def register_encoder(name: str, features: Union[str, List[str]]):
    if isinstance(features, str):
        features = [features]

    def wrap(cls):
        for feature in features:
            feature_registry = encoder_registry.get(feature, {})
            feature_registry[name] = cls
            encoder_registry[feature] = feature_registry
        return cls

    return wrap


def get_encoder_cls(feature: str, name: str) -> Type[Encoder]:
    return encoder_registry[feature][name]


def get_encoder_classes(feature: str) -> Dict[str, Type[Encoder]]:
    return encoder_registry[feature]



import contextlib
import os
import shutil

import pytest

from ludwig.api import LudwigModel
from ludwig.constants import TRAINER
from ludwig.contribs import WhyLogsCallback
from tests.integration_tests.utils import category_feature, generate_data, sequence_feature, spawn

try:
    import ray
except ImportError:
    ray = None


@contextlib.contextmanager
def ray_start(num_cpus=2):
    res = ray.init(
        num_cpus=num_cpus,
        include_dashboard=False,
        object_store_memory=150 * 1024 * 1024,
    )
    try:
        yield res
    finally:
        ray.shutdown()


def test_whylogs_callback_local(tmpdir):
    epochs = 2
    batch_size = 8
    num_examples = 32

    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 2}, reduce_input="sum")]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": epochs, "batch_size": batch_size},
    }

    data_csv = generate_data(
        input_features, output_features, os.path.join(tmpdir, "train.csv"), num_examples=num_examples
    )
    val_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "validation.csv"))
    test_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "test.csv"))

    exp_name = "whylogs_test_local"
    callback = WhyLogsCallback()

    model = LudwigModel(config, callbacks=[callback])
    model.train(training_set=data_csv, validation_set=val_csv, test_set=test_csv, experiment_name=exp_name)
    _, _ = model.predict(test_csv)

    local_training_output_dir = "output/training"
    local_prediction_output_dir = "output/prediction"

    assert os.path.isdir(local_training_output_dir) is True
    assert os.path.isdir(local_prediction_output_dir) is True


@pytest.mark.distributed
def test_whylogs_callback_dask(tmpdir):
    num_examples = 100

    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 2}, reduce_input="sum")]

    data_csv = generate_data(
        input_features, output_features, os.path.join(tmpdir, "train.csv"), num_examples=num_examples
    )
    val_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "validation.csv"))
    test_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "test.csv"))

    run_dask(input_features, output_features, data_csv, val_csv, test_csv)
    local_training_output_dir = "output/training"
    local_prediction_output_dir = "output/prediction"

    assert os.path.isdir(local_training_output_dir) is True
    assert os.path.isdir(local_prediction_output_dir) is True


@spawn
def run_dask(input_features, output_features, data_csv, val_csv, test_csv):
    epochs = 2
    batch_size = 8
    backend = {
        "type": "ray",
        "processor": {
            "parallelism": 2,
        },
        "trainer": {
            "use_gpu": False,
            "num_workers": 2,
            "resources_per_worker": {
                "CPU": 1,
                "GPU": 0,
            },
        },
    }
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": epochs, "batch_size": batch_size},
    }

    with ray_start(num_cpus=4):
        exp_name = "whylogs_test_ray"
        callback = WhyLogsCallback()
        model = LudwigModel(config, backend=backend, callbacks=[callback])
        model.train(training_set=data_csv, validation_set=val_csv, test_set=test_csv, experiment_name=exp_name)
        _, _ = model.predict(test_csv)



# flake8: noqa
# fmt: off
from ludwig.schema.metadata.parameter_metadata import ExpectedImpact, ParameterMetadata

TRAINER_METADATA = (
    {'batch_size': ParameterMetadata(ui_display_name='Batch Size',
                                     default_value_reasoning='Not too big, not too small.',
                                     example_value=None,
                                     related_parameters=['eval_batch_size'],
                                     other_information=None,
                                     description_implications="There's conflicting advice about what batch size to "
                                                              "use. "
                                                              'Using a higher batch size will achieve the highest '
                                                              "throughput and training efficiency. However, "
                                                              "there's also "
                                                              'evidence that depending on other hyperparameters, '
                                                              'a smaller '
                                                              'batch size may produce a higher quality model.',
                                     suggested_values='auto',
                                     suggested_values_reasoning='Try at least a few different batch sizes to get a '
                                                                'sense '
                                                                'of whether batch size affects model performance',
                                     commonly_used=True,
                                     expected_impact=ExpectedImpact.HIGH,
                                     literature_references=None,
                                     internal_only=False),
      'max_batch_size': ParameterMetadata(ui_display_name='Max Batch Size',
                                                  default_value_reasoning='Not typically required.',
                                                  example_value=1024,
                                                  related_parameters=['batch_size', 'increase_batch_size_on_plateau'],
                                                  description_implications='Value used to manually limit the batch '
                                                                           'sizes explored by auto batch size tuning '
                                                                           'and batch size increasing on plateau.',
                                                  suggested_values=None,
                                                  suggested_values_reasoning=None,
                                                  commonly_used=False,
                                                  expected_impact=ExpectedImpact.MEDIUM,
                                                  literature_references=None,
                                                  internal_only=False),
     'checkpoints_per_epoch': ParameterMetadata(ui_display_name='Checkpoints per epoch',
                                                default_value_reasoning='Per-epoch behavior, which scales according '
                                                                        'to the '
                                                                        'dataset size.',
                                                example_value=None,
                                                related_parameters=['train_steps', 'steps_per_checkpoint'],
                                                other_information=None,
                                                description_implications='Epoch-based evaluation (using the default: '
                                                                         '0) is '
                                                                         'an appropriate fit for tabular datasets, '
                                                                         'which '
                                                                         'are small, fit in memory, and train '
                                                                         'quickly.\n'
                                                                         '\n'
                                                                         'However, this is a poor fit for unstructured '
                                                                         'datasets, which tend to be much larger, and '
                                                                         'train more slowly due to larger models.\n'
                                                                         '\n'
                                                                         "It's important to setup evaluation such "
                                                                         "that you "
                                                                         'do not wait several hours before getting a '
                                                                         'single evaluation result. In general, '
                                                                         'it is not '
                                                                         'necessary for models to train over the '
                                                                         'entirety '
                                                                         'of a dataset, nor evaluate over the '
                                                                         'entirety of '
                                                                         'a test set, to produce useful monitoring '
                                                                         'metrics '
                                                                         'and signals to indicate model health.\n'
                                                                         '\n'
                                                                         'It is also more engaging and more valuable '
                                                                         'to '
                                                                         'ensure a frequent pulse of evaluation '
                                                                         'metrics, '
                                                                         'even if they are partial.',
                                                suggested_values='2 - 10, for larger datasets',
                                                suggested_values_reasoning='Running evaluation too frequently can be '
                                                                           'wasteful while running evaluation not '
                                                                           'frequently enough can be prohibitively '
                                                                           'uninformative. In many large-scale '
                                                                           'training '
                                                                           'runs, evaluation is often configured to '
                                                                           'run on '
                                                                           'a sub-epoch time scale, or every few '
                                                                           'thousand '
                                                                           'steps.',
                                                commonly_used=True,
                                                expected_impact=ExpectedImpact.HIGH,
                                                literature_references=None,
                                                internal_only=False),
     'decay': ParameterMetadata(ui_display_name='Decay',
                                default_value_reasoning=None,
                                example_value=None,
                                related_parameters=['decay_rate', 'decay_steps', 'learning_rate'],
                                other_information=None,
                                description_implications='It’s almost always a good idea to use a schedule. For most '
                                                         'models, try the exponential decay schedule first.\n'
                                                         '\n'
                                                         'The exponential schedule divides the learning rate by the '
                                                         'same '
                                                         'factor (%) every epoch. This means that the learning rate '
                                                         'will '
                                                         'decrease rapidly in the first few epochs, and spend more '
                                                         'epochs '
                                                         'with a lower value, but never reach exactly zero. As a rule '
                                                         'of '
                                                         'thumb, compared to training without a schedule, you can use '
                                                         'a '
                                                         'slightly higher maximum learning rate. Since the learning '
                                                         'rate '
                                                         'changes over time, the whole training is not so sensitive '
                                                         'to the '
                                                         'value picked.',
                                suggested_values=None,
                                suggested_values_reasoning='There is no go-to schedule for all models. Changing the '
                                                           'learning rate, in general, has shown to make training less '
                                                           'sensitive to the learning rate value you pick for it. So '
                                                           'using '
                                                           'a learning rate schedule can give better training '
                                                           'performance '
                                                           'and make the model converge faster',
                                commonly_used=True,
                                expected_impact=ExpectedImpact.MEDIUM,
                                literature_references=[
                                    'https://peltarion.com/knowledge-center/documentation/modeling-view/run-a-model'
                                    '/optimization-principles-(in-deep-learning)/learning-rate-schedule '
                                ],
                                internal_only=False),
     'decay_rate': ParameterMetadata(ui_display_name='Decay Rate',
                                     default_value_reasoning='4-5% decay each step is an empirically useful decary '
                                                             'rate to '
                                                             'start with.',
                                     example_value=None,
                                     related_parameters=['decay_rate, decay_steps, learning_rate'],
                                     other_information=None,
                                     description_implications='Increasing the decay rate will lower the learning rate '
                                                              'faster. This could make the model more robust to a bad '
                                                              '(too '
                                                              'high) initial learning rate, but a decay rate that is '
                                                              'too '
                                                              'high could prohibit the model from learning anything at '
                                                              'all.',
                                     suggested_values='0.9 - 0.96',
                                     suggested_values_reasoning='Since this controls exponential decay, even a small '
                                                                'decay '
                                                                'rate will still be strongly impactful.',
                                     commonly_used=False,
                                     expected_impact=ExpectedImpact.MEDIUM,
                                     literature_references=[
                                         'https://peltarion.com/knowledge-center/documentation/modeling-view/run-a'
                                         '-model'
                                         '/optimization-principles-(in-deep-learning)/learning-rate-schedule '
                                     ],
                                     internal_only=False),
     'decay_steps': ParameterMetadata(ui_display_name='Decay Steps',
                                      default_value_reasoning='This default essentially enables the `learning_rate` to '
                                                              'decay by a factor of the `decay_rate` at 10000 training '
                                                              'steps.',
                                      example_value=[5000],
                                      related_parameters=['decay_rate', 'decay_steps', 'learning_rate'],
                                      other_information=None,
                                      description_implications='By increasing the value of decay steps, '
                                                               'you are increasing '
                                                               'the number of training steps it takes to decay the '
                                                               'learning rate by a factor of `decay_rate`. In other '
                                                               'words, '
                                                               'the bigger this parameter, the slower the learning '
                                                               'rate '
                                                               'decays.',
                                      suggested_values='10000 +/- 500 at a time',
                                      suggested_values_reasoning='The decay in the learning rate is calculated as the '
                                                                 'training step divided by the `decay_steps` plus one. '
                                                                 'Then the `decay_rate` is raised to the power of this '
                                                                 'exponent which is then multiplied to the current '
                                                                 'learning rate. All this to say that the learning '
                                                                 'rate is '
                                                                 'only decayed by a factor of the set `decay_rate` '
                                                                 'when '
                                                                 'the training step reaches the `decay_steps` and then '
                                                                 'subsequently when it reaches any multiple of '
                                                                 '`decay_steps`. You can think of `decay_steps` as a '
                                                                 'rate '
                                                                 'of decay for the `decay_rate`.',
                                      commonly_used=True,
                                      expected_impact=ExpectedImpact.MEDIUM,
                                      literature_references=None,
                                      internal_only=False),
     'early_stop': ParameterMetadata(ui_display_name='Early Stop',
                                     default_value_reasoning="Deep learning models are prone to overfitting. It's "
                                                             'generally a good policy to set up some early stopping '
                                                             "criteria as it's not useful to have a model train after "
                                                             "it's "
                                                             'maximized what it can learn. 5 consecutive rounds of '
                                                             "evaluation where there hasn't been any improvement on "
                                                             "the "
                                                             'validation set (including chance) is a reasonable '
                                                             'policy to '
                                                             'start with.',
                                     example_value=None,
                                     related_parameters=['epochs', 'train_steps'],
                                     other_information=None,
                                     description_implications='Decreasing this value is a more aggressive policy. '
                                                              'Decreasing early stopping makes model training less '
                                                              'forgiving, as the model has less runway to demonstrate '
                                                              'consecutive metric improvements before the training '
                                                              'run is '
                                                              'quit. This can be efficient for pruning bad models '
                                                              'earlier, '
                                                              'but since the training process is inherently '
                                                              'non-deterministic and noisy, sometimes improvements '
                                                              'happen '
                                                              'very gradually over a long period of time.',
                                     suggested_values='5 - 10',
                                     suggested_values_reasoning="There's potentially a lot of randomness in how models "
                                                                'train, but so many consecutive rounds of no '
                                                                'improvement '
                                                                "is usually a good indicator that there's not much "
                                                                "more to "
                                                                'learn.',
                                     commonly_used=True,
                                     expected_impact=ExpectedImpact.MEDIUM,
                                     literature_references=None,
                                     internal_only=False),
     'epochs': ParameterMetadata(ui_display_name='Epochs',
                                 default_value_reasoning='A very high training length ceiling. Models will almost '
                                                         'always '
                                                         'hit early stopping criteria before hitting a 100-epoch '
                                                         'ceiling.',
                                 example_value=None,
                                 related_parameters='eval_batch_size',
                                 other_information=None,
                                 description_implications="There's conflicting advice about what batch size to use. "
                                                          'Using a higher batch size will achieve the highest '
                                                          "throughput and training efficiency. However, there's also "
                                                          'evidence that depending on other hyperparameters, a smaller '
                                                          'batch size may produce a higher quality model.',
                                 suggested_values='auto',
                                 suggested_values_reasoning='Try at least a few different batch sizes to get a sense '
                                                            'of whether batch size affects model performance',
                                 commonly_used=True,
                                 expected_impact=ExpectedImpact.HIGH,
                                 literature_references=None,
                                 internal_only=False),
 'checkpoints_per_epoch': ParameterMetadata(ui_display_name='Checkpoints per epoch',
                                            default_value_reasoning='Per-epoch behavior, which scales according to the '
                                                                    'dataset size.',
                                            example_value=None,
                                            related_parameters='train_steps, steps_per_checkpoint',
                                            other_information=None,
                                            description_implications='Epoch-based evaluation (using the default: 0) is '
                                                                     'an appropriate fit for tabular datasets, which '
                                                                     'are small, fit in memory, and train quickly.\n'
                                                                     '\n'
                                                                     'However, this is a poor fit for unstructured '
                                                                     'datasets, which tend to be much larger, and '
                                                                     'train more slowly due to larger models.\n'
                                                                     '\n'
                                                                     "It's important to setup evaluation such that you "
                                                                     'do not wait several hours before getting a '
                                                                     'single evaluation result. In general, it is not '
                                                                     'necessary for models to train over the entirety '
                                                                     'of a dataset, nor evaluate over the entirety of '
                                                                     'a test set, to produce useful monitoring metrics '
                                                                     'and signals to indicate model health.\n'
                                                                     '\n'
                                                                     'It is also more engaging and more valuable to '
                                                                     'ensure a frequent pulse of evaluation metrics, '
                                                                     'even if they are partial.',
                                            suggested_values='2 - 10, for larger datasets',
                                            suggested_values_reasoning='Running evaluation too frequently can be '
                                                                       'wasteful while running evaluation not '
                                                                       'frequently enough can be prohibitively '
                                                                       'uninformative. In many large-scale training '
                                                                       'runs, evaluation is often configured to run on '
                                                                       'a sub-epoch time scale, or every few thousand '
                                                                       'steps.',
                                            commonly_used=True,
                                            expected_impact=ExpectedImpact.HIGH,
                                            literature_references=None,
                                            internal_only=False),
 'decay': ParameterMetadata(ui_display_name='Decay',
                            default_value_reasoning=None,
                            example_value=None,
                            related_parameters='decay_rate, decay_steps, learning_rate',
                            other_information=None,
                            description_implications='It’s almost always a good idea to use a schedule. For most '
                                                     'models, try the exponential decay schedule first.\n'
                                                     '\n'
                                                     'The exponential schedule divides the learning rate by the same '
                                                     'factor (%) every epoch. This means that the learning rate will '
                                                     'decrease rapidly in the first few epochs, and spend more epochs '
                                                     'with a lower value, but never reach exactly zero. As a rule of '
                                                     'thumb, compared to training without a schedule, you can use a '
                                                     'slightly higher maximum learning rate. Since the learning rate '
                                                     'changes over time, the whole training is not so sensitive to the '
                                                     'value picked.',
                            suggested_values=None,
                            suggested_values_reasoning='There is no go-to schedule for all models. Changing the '
                                                       'learning rate, in general, has shown to make training less '
                                                       'sensitive to the learning rate value you pick for it. So using '
                                                       'a learning rate schedule can give better training performance '
                                                       'and make the model converge faster',
                            commonly_used=True,
                            expected_impact=ExpectedImpact.MEDIUM,
                            literature_references='https://peltarion.com/knowledge-center/documentation/modeling-view/run-a-model/optimization-principles-(in-deep-learning)/learning-rate-schedule',
                            internal_only=False),
 'decay_rate': ParameterMetadata(ui_display_name='Decay Rate',
                                 default_value_reasoning='4-5% decay each step is an empirically useful decary rate to '
                                                         'start with.',
                                 example_value=None,
                                 related_parameters='decay_rate, decay_steps, learning_rate',
                                 other_information=None,
                                 description_implications='Increasing the decay rate will lower the learning rate '
                                                          'faster. This could make the model more robust to a bad (too '
                                                          'high) initial learning rate, but a decay rate that is too '
                                                          'high could prohibit the model from learning anything at '
                                                          'all.',
                                 suggested_values='0.9 - 0.96',
                                 suggested_values_reasoning='Since this controls exponential decay, even a small decay '
                                                            'rate will still be strongly impactful.',
                                 commonly_used=False,
                                 expected_impact=ExpectedImpact.MEDIUM,
                                 literature_references='https://peltarion.com/knowledge-center/documentation/modeling-view/run-a-model/optimization-principles-(in-deep-learning)/learning-rate-schedule',
                                 internal_only=False),
 'decay_steps': ParameterMetadata(ui_display_name='Decay Steps',
                                  default_value_reasoning='This default essentially enables the `learning_rate` to '
                                                          'decay by a factor of the `decay_rate` at 10000 training '
                                                          'steps.',
                                  example_value=5000,
                                  related_parameters='decay_rate, decay_steps, learning_rate',
                                  other_information=None,
                                  description_implications='By increasing the value of decay steps, you are increasing '
                                                           'the number of training steps it takes to decay the '
                                                           'learning rate by a factor of `decay_rate`. In other words, '
                                                           'the bigger this parameter, the slower the learning rate '
                                                           'decays.',
                                  suggested_values='10000 +/- 500 at a time',
                                  suggested_values_reasoning='The decay in the learning rate is calculated as the '
                                                             'training step divided by the `decay_steps` plus one. '
                                                             'Then the `decay_rate` is raised to the power of this '
                                                             'exponent which is then multiplied to the current '
                                                             'learning rate. All this to say that the learning rate is '
                                                             'only decayed by a factor of the set `decay_rate` when '
                                                             'the training step reaches the `decay_steps` and then '
                                                             'subsequently when it reaches any multiple of '
                                                             '`decay_steps`. You can think of `decay_steps` as a rate '
                                                             'of decay for the `decay_rate`.',
                                  commonly_used=True,
                                  expected_impact=ExpectedImpact.MEDIUM,
                                  literature_references=None,
                                  internal_only=False),
 'early_stop': ParameterMetadata(ui_display_name='Early Stop',
                                 default_value_reasoning="Deep learning models are prone to overfitting. It's "
                                                         'generally a good policy to set up some early stopping '
                                                         "criteria as it's not useful to have a model train after it's "
                                                         'maximized what it can learn. 5 consecutive rounds of '
                                                         "evaluation where there hasn't been any improvement on the "
                                                         'validation set (including chance) is a reasonable policy to '
                                                         'start with.',
                                 example_value=None,
                                 related_parameters='epochs, train_steps',
                                 other_information=None,
                                 description_implications='Decreasing this value is a more aggressive policy. '
                                                          'Decreasing early stopping makes model training less '
                                                          'forgiving, as the model has less runway to demonstrate '
                                                          'consecutive metric improvements before the training run is '
                                                          'quit. This can be efficient for pruning bad models earlier, '
                                                          'but since the training process is inherently '
                                                          'non-deterministic and noisy, sometimes improvements happen '
                                                          'very gradually over a long period of time.',
                                 suggested_values='5 - 10',
                                 suggested_values_reasoning="There's potentially a lot of randomness in how models "
                                                            'train, but so many consecutive rounds of no improvement '
                                                            "is usually a good indicator that there's not much more to "
                                                            'learn.',
                                 commonly_used=True,
                                 expected_impact=ExpectedImpact.MEDIUM,
                                 literature_references=None,
                                 internal_only=False),
 'epochs': ParameterMetadata(ui_display_name='Epochs',
                             default_value_reasoning='A very high training length ceiling. Models will almost always '
                                                     'hit early stopping criteria before hitting a 100-epoch ceiling.',
                             example_value=None,
                             related_parameters='train_steps',
                             other_information=None,
                             description_implications='Decreasing this will shorten the overall runway for training '
                                                      'the model.',
                             suggested_values='0 (and use train_steps), or 100',
                             suggested_values_reasoning="Usually it's sensible to leave this very high and rely on a "
                                                        'solid early stopping policy to dictate when the model should '
                                                        'stop training. Some models and hyperparameter configurations '
                                                        'require many epochs through the dataset to converge while '
                                                        'others converge before a single epoch through the data.',
                             commonly_used=False,
                             expected_impact=ExpectedImpact.HIGH,
                             literature_references=None,
                             internal_only=False),
 'eval_batch_size': ParameterMetadata(ui_display_name='Evaluation Batch Size',
                                      default_value_reasoning='Use the entire batch for model evaluation unless '
                                                              'otherwise specified',
                                      example_value=512,
                                      related_parameters='batch_size',
                                      other_information='Should only set the batch_size to a level that you can fit in '
                                                        'memory',
                                      description_implications='By increasing the `eval_batch_size` past the '
                                                               '`batch_size` parameter set value, you allow for more '
                                                               'parallelism in the batch evaluation step. For example, '
                                                               'if you have to evaluate the model on a test set of '
                                                               'size 1000, it is faster to evaluate two times with two '
                                                               'batches of size 500 as opposed to ten times with ten '
                                                               'batches of 100. Setting this parameter higher without '
                                                               'maxing out memory limits will speed up the model '
                                                               'training process overall.',
                                      suggested_values=(256, 512, 1024),
                                      suggested_values_reasoning='By observing memory consumption on training jobs, '
                                                                 'you can get a sense of how much extra memory is '
                                                                 'available for increasing this value. A good rule of '
                                                                 'thumb can be experimentally doubling the eval batch '
                                                                 'size if you do not have insight into memory usage.',
                                      commonly_used=False,
                                      expected_impact=ExpectedImpact.MEDIUM,
                                      literature_references=None,
                                      internal_only=False),
 'evaluate_training_set': ParameterMetadata(ui_display_name='Evaluate Training Set',
                                            default_value_reasoning='It could be useful to monitor evaluation metrics '
                                                                    'on the training set, as a secondary validation '
                                                                    'set.',
                                            example_value=None,
                                            related_parameters=None,
                                            other_information=None,
                                            description_implications='Running evaluation on the full training set, '
                                                                     'when your training set is large, can be a huge '
                                                                     'computational cost. Turning off training set '
                                                                     'evaluation will lead to significant gains in '
                                                                     'training throughput and efficiency. For small '
                                                                     'datasets that train and evaluate quickly, the '
                                                                     'choice is trivial.',
                                            suggested_values=False,
                                            suggested_values_reasoning='Running full-scale evaluation on the full '
                                                                       "training set doesn't usually provide any "
                                                                       'useful information over the validation '
                                                                       'dataset. Even with this set to False, '
                                                                       'continuous training loss metrics are still '
                                                                       'computed, so it will still be easy to spot '
                                                                       'signs of overfitting like when the '
                                                                       'training-validation loss curves diverge.',
                                            commonly_used=True,
                                            expected_impact=ExpectedImpact.HIGH,
                                            literature_references=None,
                                            internal_only=False),
 'gradient_clipping': ParameterMetadata(ui_display_name='Gradient Clipping',
                                        default_value_reasoning='A conservative cap on the maximum gradient size to '
                                                                'apply over a single training step.',
                                        example_value=None,
                                        related_parameters=None,
                                        other_information=None,
                                        description_implications='Gradient clipping is a technique to prevent '
                                                                 'exploding gradients in very deep networks. '
                                                                 'Increasing gradient clipping can help with model '
                                                                 'training loss curve stability, but it can also make '
                                                                 'training less efficient as weight at each training '
                                                                 'step is capped.',
                                        suggested_values=None,
                                        suggested_values_reasoning="It's usually sensible to have some conservative "
                                                                   'notion of gradient clipping to make modeling '
                                                                   'robust to a particularly bad or noisy batch of '
                                                                   'examples.',
                                        commonly_used=False,
                                        expected_impact=ExpectedImpact.MEDIUM,
                                        literature_references=None,
                                        internal_only=False),
 'increase_batch_size_eval_metric': ParameterMetadata(ui_display_name='Batch Size Increase: Evaluation Metric',
                                                      default_value_reasoning=None,
                                                      example_value=None,
                                                      related_parameters=None,
                                                      other_information=None,
                                                      description_implications=None,
                                                      suggested_values=None,
                                                      suggested_values_reasoning=None,
                                                      commonly_used=False,
                                                      expected_impact=ExpectedImpact.MEDIUM,
                                                      literature_references=None,
                                                      internal_only=False),
 'increase_batch_size_eval_split': ParameterMetadata(ui_display_name='Batch Size Increase: Evaluation Split',
                                                     default_value_reasoning=None,
                                                     example_value=None,
                                                     related_parameters=None,
                                                     other_information=None,
                                                     description_implications=None,
                                                     suggested_values=None,
                                                     suggested_values_reasoning=None,
                                                     commonly_used=False,
                                                     expected_impact=ExpectedImpact.MEDIUM,
                                                     literature_references=None,
                                                     internal_only=False),
 'increase_batch_size_on_plateau': ParameterMetadata(ui_display_name='Batch Size Increase On Plateau',
                                                     default_value_reasoning=None,
                                                     example_value=None,
                                                     related_parameters=None,
                                                     other_information=None,
                                                     description_implications=None,
                                                     suggested_values=None,
                                                     suggested_values_reasoning=None,
                                                     commonly_used=False,
                                                     expected_impact=ExpectedImpact.MEDIUM,
                                                     literature_references=None,
                                                     internal_only=False),
 'increase_batch_size_on_plateau_max': ParameterMetadata(ui_display_name='Batch Size Increase On Plateau: Cap',
                                                         default_value_reasoning=None,
                                                         example_value=None,
                                                         related_parameters=None,
                                                         other_information=None,
                                                         description_implications=None,
                                                         suggested_values=None,
                                                         suggested_values_reasoning=None,
                                                         commonly_used=False,
                                                         expected_impact=ExpectedImpact.MEDIUM,
                                                         literature_references=None,
                                                         internal_only=False),
     'increase_batch_size_on_plateau_patience': ParameterMetadata(ui_display_name='Batch Size Increase On Plateau: '
                                                                                  'Patience',
                                                                  default_value_reasoning=None,
                                                                  example_value=None,
                                                                  related_parameters=None,
                                                                  other_information=None,
                                                                  description_implications=None,
                                                                  suggested_values=None,
                                                                  suggested_values_reasoning=None,
                                                                  commonly_used=False,
                                                                  expected_impact=ExpectedImpact.MEDIUM,
                                                                  literature_references=None,
                                                                  internal_only=False),
     'increase_batch_size_on_plateau_rate': ParameterMetadata(ui_display_name='Batch Size Increase On Plateau: Rate',
                                                              default_value_reasoning=None,
                                                              example_value=None,
                                                              related_parameters=None,
                                                              other_information=None,
                                                              description_implications=None,
                                                              suggested_values=None,
                                                              suggested_values_reasoning=None,
                                                              commonly_used=False,
                                                              expected_impact=ExpectedImpact.MEDIUM,
                                                              literature_references=None,
                                                              internal_only=False),
 'increase_batch_size_on_plateau_rate': ParameterMetadata(ui_display_name='Batch Size Increase On Plateau: Rate',
                                                          default_value_reasoning=None,
                                                          example_value=None,
                                                          related_parameters=None,
                                                          other_information=None,
                                                          description_implications=None,
                                                          suggested_values=None,
                                                          suggested_values_reasoning=None,
                                                          commonly_used=False,
                                                          expected_impact=ExpectedImpact.MEDIUM,
                                                          literature_references=None,
                                                          internal_only=False),
 'learning_rate': ParameterMetadata(ui_display_name='Learning Rate',
                                    default_value_reasoning='Middle of the road learning rate to start with.',
                                    example_value=None,
                                    related_parameters=None,
                                    other_information=None,
                                    description_implications='The learning rate is a hyperparameter that controls how '
                                                             'much to change the model in response to the estimated '
                                                             'error each time the model weights are updated. '
                                                             'Increasing the learning rate may decrease learning curve '
                                                             'stability but also increase learning speed and '
                                                             'efficiency, leading to faster model convergence. '
                                                             'Decreasing the learning rate can help stabilize learning '
                                                             'curves at the cost of slower time to convergence.',
                                    suggested_values='0.00001 - 0.1',
                                    suggested_values_reasoning='Tabular models trained from scratch typically use '
                                                               'learning rates around 1e-3 while learning rates for '
                                                               'pre-trained models should be much smaller, typically '
                                                               'around 1e-5, which is important to mitigate '
                                                               'catastrophic forgetting. To make the model more robust '
                                                               'to any specific choice of learning rate, consider '
                                                               'turning enabling learning rate exponential decay.',
                                    commonly_used=True,
                                    expected_impact=ExpectedImpact.HIGH,
                                    literature_references=None,
                                    internal_only=False),
 'learning_rate_scaling': ParameterMetadata(ui_display_name='Learning Rate Scaling',
                                            default_value_reasoning='Traditionally the learning rate is scaled '
                                                                    'linearly with the number of workers to reflect '
                                                                    'the proportion by which the effective batch size '
                                                                    'is increased.',
                                            example_value=None,
                                            related_parameters=None,
                                            other_information=None,
                                            description_implications='Traditionally the learning rate is scaled '
                                                                     'linearly with the number of workers to reflect '
                                                                     'the proportion by which the effective batch size '
                                                                     'is increased. For very large batch sizes, a '
                                                                     'softer square-root scale can sometimes lead to '
                                                                     'better model performance. If the learning rate '
                                                                     'is hand-tuned for a given number of workers, '
                                                                     'setting this value to constant can be used to '
                                                                     'disable scale-up.',
                                            suggested_values='linear or sqrt',
                                            suggested_values_reasoning='Traditionally the learning rate is scaled '
                                                                       'linearly with the number of workers to reflect '
                                                                       'the proportion by which the effective batch '
                                                                       'size is increased. For very large batch sizes, '
                                                                       'a softer square-root scale can sometimes lead '
                                                                       'to better model performance. If the learning '
                                                                       'rate is hand-tuned for a given number of '
                                                                       'workers, setting this value to constant can be '
                                                                       'used to disable scale-up.',
                                            commonly_used=False,
                                            expected_impact=ExpectedImpact.MEDIUM,
                                            literature_references=None,
                                            internal_only=False),
 'learning_rate_warmup_epochs': ParameterMetadata(ui_display_name='Learning Rate Warmup Epochs',
                                                  default_value_reasoning='The randomness of how weights are '
                                                                          'initialized can result in strange, noisy '
                                                                          'gradient updates during the begining of '
                                                                          'your training run. Learning rate warmup has '
                                                                          'the benefit of slowly starting to tune '
                                                                          'things like attention mechanisms in your '
                                                                          'network, that may be prone to bad initial '
                                                                          'conditions.',
                                                  example_value=None,
                                                  related_parameters=None,
                                                  other_information=None,
                                                  description_implications='Learning rate warmup sets a very low '
                                                                           'learning rate for a set number of training '
                                                                           'steps (warmup steps). After your warmup '
                                                                           'steps you use your "regular" learning rate '
                                                                           'or learning rate scheduler. You can also '
                                                                           'gradually increase your learning rate over '
                                                                           'the number of warmup steps.',
                                                  suggested_values=True,
                                                  suggested_values_reasoning="You don't want to warm up for too long, "
                                                                             'as after the model is starting to hill '
                                                                             'climb, you want to use the full weight '
                                                                             'of the learning rate to descend into '
                                                                             'good loss minima.',
                                                  commonly_used=False,
                                                  expected_impact=ExpectedImpact.MEDIUM,
                                                  literature_references='https://datascience.stackexchange.com/questions/55991/in-the-context-of-deep-learning-what-is-training-warmup-steps',
                                                  internal_only=False),
 'optimizer': ParameterMetadata(ui_display_name='Optimizer',
                                default_value_reasoning='First try Adam because it is more likely to return good '
                                                        'results without an advanced fine tuning.',
                                example_value=None,
                                related_parameters=None,
                                other_information=None,
                                description_implications='Choosing a good optimizer for your machine learning project '
                                                         'can be overwhelming. Popular deep learning libraries such as '
                                                         'PyTorch or TensorFLow offer a broad selection of different '
                                                         'optimizers — each with its own strengths and weaknesses. '
                                                         'However, picking the wrong optimizer can have a substantial '
                                                         'negative impact on the performance of your machine learning '
                                                         'model [1][2]. This makes optimizers a critical design choice '
                                                         'in the process of building, testing, and deploying your '
                                                         'machine learning model.',
                                suggested_values='adam, adamw',
                                suggested_values_reasoning='As a rule of thumb: If you have the resources to find a '
                                                           'good learning rate schedule, SGD with momentum is a solid '
                                                           'choice. If you are in need of quick results without '
                                                           'extensive hypertuning, tend towards adaptive gradient '
                                                           'methods like adam or adamw.',
                                commonly_used=True,
                                expected_impact=ExpectedImpact.HIGH,
                                literature_references='https://www.youtube.com/watch?v=mdKjMPmcWjY',
                                internal_only=False),
 'reduce_learning_rate_eval_metric': ParameterMetadata(ui_display_name='Reduce Learning Rate Eval Metric',
                                                       default_value_reasoning=None,
                                                       example_value=None,
                                                       related_parameters=None,
                                                       other_information=None,
                                                       description_implications=None,
                                                       suggested_values=None,
                                                       suggested_values_reasoning=None,
                                                       commonly_used=False,
                                                       expected_impact=ExpectedImpact.LOW,
                                                       literature_references=None,
                                                       internal_only=False),
 'reduce_learning_rate_eval_split': ParameterMetadata(ui_display_name='Reduce Learning Rate Eval Split',
                                                      default_value_reasoning=None,
                                                      example_value=None,
                                                      related_parameters=None,
                                                      other_information=None,
                                                      description_implications=None,
                                                      suggested_values=None,
                                                      suggested_values_reasoning=None,
                                                      commonly_used=False,
                                                      expected_impact=ExpectedImpact.LOW,
                                                      literature_references=None,
                                                      internal_only=False),
 'reduce_learning_rate_on_plateau': ParameterMetadata(ui_display_name='Reduce Learning Rate On Plateau',
                                                      default_value_reasoning=None,
                                                      example_value=None,
                                                      related_parameters=None,
                                                      other_information=None,
                                                      description_implications=None,
                                                      suggested_values=None,
                                                      suggested_values_reasoning=None,
                                                      commonly_used=False,
                                                      expected_impact=ExpectedImpact.LOW,
                                                      literature_references=None,
                                                      internal_only=False),
 'reduce_learning_rate_on_plateau_patience': ParameterMetadata(ui_display_name='Reduce Learning Rate on Plateau '
                                                                               'Patience',
                                                               default_value_reasoning=None,
                                                               example_value=None,
                                                               related_parameters=None,
                                                               other_information=None,
                                                               description_implications=None,
                                                               suggested_values=None,
                                                               suggested_values_reasoning=None,
                                                               commonly_used=False,
                                                               expected_impact=ExpectedImpact.LOW,
                                                               literature_references=None,
                                                               internal_only=False),
 'reduce_learning_rate_on_plateau_rate': ParameterMetadata(ui_display_name='Reduce Learning Rate On Plateau Rate',
                                                           default_value_reasoning=None,
                                                           example_value=None,
                                                           related_parameters=None,
                                                           other_information=None,
                                                           description_implications=None,
                                                           suggested_values=None,
                                                           suggested_values_reasoning=None,
                                                           commonly_used=False,
                                                           expected_impact=ExpectedImpact.LOW,
                                                           literature_references=None,
                                                           internal_only=False),
 'regularization_lambda': ParameterMetadata(ui_display_name='Regularization Lambda',
                                            default_value_reasoning='How to tune the overall impact of the '
                                                                    'regularization term by multiplying its value by a '
                                                                    'scalar known as lambda (also called the '
                                                                    'regularization rate).',
                                            example_value=None,
                                            related_parameters='regularization_type',
                                            other_information=None,
                                            description_implications='When choosing a lambda value, the goal is to '
                                                                     'strike the right balance between simplicity and '
                                                                     'training-data fit:\n'
                                                                     '\n'
                                                                     'If your lambda value is too high, your model '
                                                                     'will be simple, but you run the risk of '
                                                                     "underfitting your data. Your model won't learn "
                                                                     'enough about the training data to make useful '
                                                                     'predictions.\n'
                                                                     '\n'
                                                                     'If your lambda value is too low, your model will '
                                                                     'be more complex, and you run the risk of '
                                                                     'overfitting your data. Your model will learn too '
                                                                     'much about the particularities of the training '
                                                                     "data, and won't be able to generalize to new "
                                                                     'data. The ideal value of lambda produces a model '
                                                                     'that generalizes well to new, previously unseen '
                                                                     'data. Unfortunately, that ideal value of lambda '
                                                                     "is data-dependent, so you'll need to do some "
                                                                     'tuning.',
                                            suggested_values=0.1,
                                            suggested_values_reasoning='The most common type of regularization is L2, '
                                                                       'also called simply “weight decay,” with values '
                                                                       'often on a logarithmic scale between 0 and '
                                                                       '0.1, such as 0.1, 0.001, 0.0001, etc.',
                                            commonly_used=True,
                                            expected_impact=ExpectedImpact.HIGH,
                                            literature_references='https://developers.google.com/machine-learning/crash-course/regularization-for-simplicity/lambda',
                                            internal_only=False),
 'regularization_type': ParameterMetadata(ui_display_name='Regularization Type',
                                          default_value_reasoning='L2 is a sufficiently good regularization to start '
                                                                  'with.',
                                          example_value=None,
                                          related_parameters='regularization_lambda',
                                          other_information=None,
                                          description_implications='L1 regularization penalizes the sum of absolute '
                                                                   'values of the weights, whereas L2 regularization '
                                                                   'penalizes the sum of squares of the weights. \n'
                                                                   'The L1 regularization solution is sparse. The L2 '
                                                                   'regularization solution is non-sparse.\n'
                                                                   'L2 regularization doesn’t perform feature '
                                                                   'selection, since weights are only reduced to '
                                                                   'values near 0 instead of 0. L1 regularization has '
                                                                   'built-in feature selection.\n'
                                                                   'L1 regularization is robust to outliers, L2 '
                                                                   'regularization is not.',
                                          suggested_values='L2',
                                          suggested_values_reasoning=None,
                                          commonly_used=True,
                                          expected_impact=ExpectedImpact.HIGH,
                                          literature_references='https://neptune.ai/blog/fighting-overfitting-with-l1-or-l2-regularization#:~:text=The%20differences%20between%20L1%20and,regularization%20solution%20is%20non%2Dsparse.',
                                          internal_only=False),
 'should_shuffle': ParameterMetadata(ui_display_name='Should Shuffle',
                                     default_value_reasoning="In general, it's a good idea to mix up data on each "
                                                             'batch so that the neural network gets the broadest '
                                                             'exposure to the dataset.',
                                     example_value=None,
                                     related_parameters=None,
                                     other_information=None,
                                     description_implications='Turning off mini-batch shuffling can make training '
                                                              'faster, but it may lead to worse performance overall as '
                                                              'shuffling helps mitigate overfitting.',
                                     suggested_values=True,
                                     suggested_values_reasoning='One of the most powerful things about neural networks '
                                                                'is that they can be very complex functions, allowing '
                                                                'one to learn very complex relationships between your '
                                                                'input and output data. These relationships can '
                                                                'include things you would never expect, such as the '
                                                                'order in which data is fed in per epoch. If the order '
                                                                'of data within each epoch is the same, then the model '
                                                                'may use this as a way of reducing the training error, '
                                                                'which is a sort of overfitting.',
                                     commonly_used=False,
                                     expected_impact='High',
                                     literature_references='https://stats.stackexchange.com/questions/245502/why-should-we-shuffle-data-while-training-a-neural-network#:~:text=it%20helps%20the%20training%20converge,the%20order%20of%20the%20training',
                                     internal_only=False),
 'staircase': ParameterMetadata(ui_display_name='Staircase',
                                default_value_reasoning='Extra knob for decaying the learning rate in a different way.',
                                example_value=None,
                                related_parameters=None,
                                other_information=None,
                                description_implications='An excessively aggressive decay results in optimizers never '
                                                         'reaching the minima, whereas a slow decay leads to chaotic '
                                                         'updates without significant improvement. Discrete learning '
                                                         'rate decay is another knob to help tune a balance.',
                                suggested_values=False,
                                suggested_values_reasoning='We have not found strong evidence that discretely decaying '
                                                           'the learning rate is superior to doing so continuously.',
                                commonly_used=False,
                                expected_impact=ExpectedImpact.LOW,
                                literature_references='https://neptune.ai/blog/how-to-choose-a-learning-rate-scheduler',
                                internal_only=False),
 'steps_per_checkpoint': ParameterMetadata(ui_display_name='Steps Per Checkpoint',
                                           default_value_reasoning='By default, we evaluate once per epoch, which '
                                                                   'scales according to the dataset size.',
                                           example_value=None,
                                           related_parameters='checkpoints_per_epoch',
                                           other_information=None,
                                           description_implications='Epoch-based evaluation (using the default: 0) is '
                                                                    'an appropriate fit for tabular datasets, which '
                                                                    'are small, fit in memory, and train quickly.\n'
                                                                    '\n'
                                                                    'However, this is a poor fit for unstructured '
                                                                    'datasets, which tend to be much larger, and train '
                                                                    'more slowly due to larger models.\n'
                                                                    '\n'
                                                                    "It's important to setup evaluation such that you "
                                                                    'do not wait several hours before getting a single '
                                                                    'evaluation result. In general, it is not '
                                                                    'necessary for models to train over the entirety '
                                                                    'of a dataset, nor evaluate over the entirety of a '
                                                                    'test set, to produce useful monitoring metrics '
                                                                    'and signals to indicate model health.\n'
                                                                    '\n'
                                                                    'It is also more engaging and more valuable to '
                                                                    'ensure a frequent pulse of evaluation metrics, '
                                                                    'even if they are partial.',
                                           suggested_values='O(1k) for larger datasets',
                                           suggested_values_reasoning='Running evaluation too frequently can be '
                                                                      'wasteful while running evaluation not '
                                                                      'frequently enough can be prohibitively '
                                                                      'uninformative. In many large-scale training '
                                                                      'runs, evaluation is often configured to run on '
                                                                      'a sub-epoch time scale, or every few thousand '
                                                                      'steps.',
                                           commonly_used=True,
                                           expected_impact=ExpectedImpact.HIGH,
                                           literature_references=None,
                                           internal_only=False),
 'train_steps': ParameterMetadata(ui_display_name='Train Steps',
                                  default_value_reasoning='This defaults to `epochs`, which is a very high training '
                                                          'length ceiling. Models will almost always hit early '
                                                          'stopping criteria before reaching the absolute end of the '
                                                          'training runway.',
                                  example_value=None,
                                  related_parameters='epochs',
                                  other_information=None,
                                  description_implications='Decreasing this will shorten the overall runway for '
                                                           'training the model.',
                                  suggested_values='0 (and use epochs), or 1000000, 1 for debugging',
                                  suggested_values_reasoning="Usually it's sensible to leave this very high and rely "
                                                             'on a solid early stopping policy to dictate when the '
                                                             'model should stop training. Some models and '
                                                             'hyperparameter configurations require many epochs '
                                                             'through the dataset to converge while others converge '
                                                             'before a single epoch through the data.',
                                  commonly_used=False,
                                  expected_impact=ExpectedImpact.HIGH,
                                  literature_references=None,
                                  internal_only=False),
 'validation_field': ParameterMetadata(ui_display_name='Validation Field',
                                       default_value_reasoning='Concrete evaluation metrics are usually better than '
                                                               'loss, the penalty for a bad prediction, which is only '
                                                               'a proxy for prediction correctness.',
                                       example_value=None,
                                       related_parameters='validation_field, validation_metric',
                                       other_information=None,
                                       description_implications='This parameter affects 1) what the early stopping '
                                                                'policy looks at to determine when to early stop and '
                                                                '2) hyperparameter optimization for determining the '
                                                                'best trial.',
                                       suggested_values='default behavior',
                                       suggested_values_reasoning=None,
                                       commonly_used=False,
                                       expected_impact=ExpectedImpact.HIGH,
                                       literature_references=None,
                                       internal_only=False),
 'validation_metric': ParameterMetadata(ui_display_name='Validation Metric',
                                        default_value_reasoning=None,
                                        example_value=None,
                                        related_parameters='validation_field, validation_metric',
                                        other_information=None,
                                        description_implications='This parameter affects 1) what the early stopping '
                                                                 'policy looks at to determine when to early stop and '
                                                                 '2) hyperparameter optimization for determining the '
                                                                 'best trial.',
                                        suggested_values='default behavior',
                                        suggested_values_reasoning=None,
                                        commonly_used=False,
                                        expected_impact='NONE',
                                        literature_references=None,
                                        internal_only=False)}
)



from dataclasses import field

from marshmallow import fields, ValidationError

import ludwig.schema.utils as schema_utils
from ludwig.schema.features.utils import input_config_registry, output_config_registry


def DefaultsDataclassField(feature_type: str):
    """Custom dataclass field that when used inside a dataclass will allow the user to specify a nested default
    config for a specific feature type.

    Returns: Initialized dataclass field that converts an untyped dict with params to a defaults config.
    """

    class DefaultMarshmallowField(fields.Field):
        """Custom marshmallow field that deserializes a dict for a valid defaults config from the feature_registry
        and creates a corresponding JSON schema for external usage."""

        def _deserialize(self, value, attr, data, **kwargs):
            if value is None:
                return None
            if isinstance(value, dict):
                input_feature_class = input_config_registry[feature_type]
                output_feature_class = output_config_registry.get(feature_type, None)
                try:
                    input_schema = input_feature_class.Schema().load(value)
                    if output_feature_class:
                        output_schema = output_feature_class.Schema().load(value)
                        combined = input_schema + output_schema
                    else:
                        combined = input_schema
                    return combined
                except (TypeError, ValidationError) as error:
                    raise ValidationError(f"Invalid params: {value}, see `{attr}` definition. Error: {error}")
            raise ValidationError(f"Invalid params: {value}")

        @staticmethod
        def _jsonschema_type_mapping():
            input_feature_cls = input_config_registry.get(feature_type)
            output_feature_cls = output_config_registry.get(feature_type, None)
            input_props = schema_utils.unload_jsonschema_from_marshmallow_class(input_feature_cls)["properties"]
            if output_feature_cls:
                output_props = schema_utils.unload_jsonschema_from_marshmallow_class(output_feature_cls)["properties"]
                combined_props = {**output_props, **input_props}
            else:
                combined_props = input_props
            return {
                "type": "object",
                "properties": combined_props,
                "title": "defaults_options",
            }

    try:
        input_cls = input_config_registry[feature_type]
        output_cls = output_config_registry.get(feature_type, None)
        dump_default = input_cls.Schema().dump({"type": feature_type})
        if output_cls:
            output_dump = output_cls.Schema().dump({"type": feature_type})
            dump_default = {**output_dump, **dump_default}

        load_default = input_cls.Schema().load({"type": feature_type})
        if output_cls:
            output_load = output_cls.Schema().load({"type": feature_type})
            for k in dump_default.keys():
                if getattr(load_default, k, -1) == -1:
                    setattr(load_default, k, getattr(output_load, k))
        return field(
            metadata={
                "marshmallow_field": DefaultMarshmallowField(
                    allow_none=False,
                    dump_default=dump_default,
                    load_default=load_default,
                )
            },
            default_factory=lambda: load_default,
        )
    except Exception as e:
        raise ValidationError(f"Unsupported feature type: {feature_type}. See input_type_registry. " f"Details: {e}")



import pytest

from ludwig.schema import validate_config
from tests.integration_tests.utils import binary_feature, category_feature


def test_config_preprocessing():
    input_features = [category_feature(), category_feature()]
    output_features = [binary_feature()]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "preprocessing": {
            "split": {
                "type": "random",
                "probabilities": [0.6, 0.2, 0.2],
            },
            "oversample_minority": 0.4,
        },
    }

    validate_config(config)

    config["preprocessing"]["fake_parameter"] = True

    with pytest.raises(Exception):
        validate_config(config)



from typing import Optional

from marshmallow_dataclass import dataclass

from ludwig.schema import utils as schema_utils
from ludwig.schema.combiners.base import BaseCombinerConfig


@dataclass
class TabNetCombinerConfig(BaseCombinerConfig):
    """Parameters for tabnet combiner."""

    size: int = schema_utils.PositiveInteger(default=32, description="`N_a` in the paper.")

    output_size: int = schema_utils.PositiveInteger(
        default=128, description="Output size of a fully connected layer. `N_d` in the paper"
    )

    num_steps: int = schema_utils.NonNegativeInteger(
        default=3,
        description=(
            "Number of steps / repetitions of the the attentive transformer and feature transformer computations. "
            "`N_steps` in the paper"
        ),
    )

    num_total_blocks: int = schema_utils.NonNegativeInteger(
        default=4, description="Total number of feature transformer block at each step"
    )

    num_shared_blocks: int = schema_utils.NonNegativeInteger(
        default=2, description="Number of shared feature transformer blocks across the steps"
    )

    relaxation_factor: float = schema_utils.FloatRange(
        default=1.5,
        description=(
            "Factor that influences how many times a feature should be used across the steps of computation. a value of"
            " 1 implies it each feature should be use once, a higher value allows for multiple usages. `gamma` in the "
            "paper"
        ),
    )

    bn_epsilon: float = schema_utils.FloatRange(
        default=1e-3, description="Epsilon to be added to the batch norm denominator."
    )

    bn_momentum: float = schema_utils.FloatRange(
        default=0.05, description="Momentum of the batch norm. 1 - `m_B` from the TabNet paper."
    )

    bn_virtual_bs: Optional[int] = schema_utils.PositiveInteger(
        default=1024,
        allow_none=True,
        description=(
            "Size of the virtual batch size used by ghost batch norm. If null, regular batch norm is used instead. "
            "`B_v` from the TabNet paper"
        ),
    )

    sparsity: float = schema_utils.FloatRange(
        default=1e-4, description="Multiplier of the sparsity inducing loss. `lambda_sparse` in the paper"
    )

    entmax_mode: str = schema_utils.StringOptions(
        ["entmax15", "sparsemax", "constant", "adaptive"], default="sparsemax", description=""
    )

    entmax_alpha: float = schema_utils.FloatRange(
        default=1.5, min=1, max=2, description=""
    )  # 1 corresponds to softmax, 2 is sparsemax.

    dropout: float = schema_utils.FloatRange(
        default=0.05, min=0, max=1, description="Dropout rate for the transformer block."
    )



from marshmallow_dataclass import dataclass

from ludwig.constants import CATEGORY, DROP_ROW, MISSING_VALUE_STRATEGY_OPTIONS
from ludwig.schema import utils as schema_utils
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import register_preprocessor
from ludwig.schema.metadata.preprocessing_metadata import PREPROCESSING_METADATA
from ludwig.utils import strings_utils


@register_preprocessor(CATEGORY)
@dataclass
class CategoryPreprocessingConfig(BasePreprocessingConfig):
    """CategoryPreprocessingConfig is a dataclass that configures the parameters used for a category input
    feature."""

    missing_value_strategy: str = schema_utils.StringOptions(
        MISSING_VALUE_STRATEGY_OPTIONS,
        default="fill_with_const",
        allow_none=False,
        description="What strategy to follow when there's a missing value in a category column",
    )

    fill_value: str = schema_utils.String(
        default=strings_utils.UNKNOWN_SYMBOL,
        allow_none=False,
        description="The value to replace missing values with in case the missing_value_strategy is fill_with_const",
    )

    computed_fill_value: str = schema_utils.String(
        default=strings_utils.UNKNOWN_SYMBOL,
        allow_none=False,
        description="The internally computed fill value to replace missing values with in case the "
        "missing_value_strategy is fill_with_mode or fill_with_mean",
        parameter_metadata=PREPROCESSING_METADATA["computed_fill_value"],
    )

    lowercase: bool = schema_utils.Boolean(
        default=False,
        description="Whether the string has to be lowercased before being handled by the tokenizer.",
    )

    most_common: int = schema_utils.PositiveInteger(
        default=10000,
        allow_none=True,
        description="The maximum number of most common tokens to be considered. if the data contains more than this "
        "amount, the most infrequent tokens will be treated as unknown.",
    )


@register_preprocessor("category_output")
@dataclass
class CategoryOutputPreprocessingConfig(CategoryPreprocessingConfig):

    missing_value_strategy: str = schema_utils.StringOptions(
        MISSING_VALUE_STRATEGY_OPTIONS,
        default=DROP_ROW,
        allow_none=False,
        description="What strategy to follow when there's a missing value in a category output feature",
    )



import argparse
import importlib
import logging
import os
import shutil
from typing import Any, Dict, Union

from ludwig.api import LudwigModel
from ludwig.benchmarking.utils import delete_model_checkpoints, export_artifacts, load_from_module
from ludwig.contrib import add_contrib_callback_args
from ludwig.utils.data_utils import load_yaml

os.environ["TOKENIZERS_PARALLELISM"] = "false"
logger = logging.getLogger(__name__)


def setup_experiment(experiment: Dict[str, str]) -> Dict[Any, Any]:
    """Set up the backend and load the Ludwig config.

    experiment: dictionary containing the dataset name, config path, and experiment name.
    Returns a Ludwig config.
    """
    shutil.rmtree(os.path.join(experiment["dataset_name"]), ignore_errors=True)
    model_config = load_yaml(experiment["config_path"])
    model_config["backend"] = {}
    model_config["backend"]["type"] = "local"
    return model_config


def benchmark_one_local(experiment: Dict[str, str], export_artifacts_dict: Dict[str, Union[str, bool]]) -> None:
    """Run a Ludwig exepriment and track metrics given a dataset name.

    experiment: dictionary containing the dataset name, config path, and experiment name.
    export_artifacts_dict: dictionary containing an export boolean flag and a path to export to.
    """
    logger.info(f"\nRunning experiment *{experiment['experiment_name']}* on dataset *{experiment['dataset_name']}*")

    # configuring backend and paths
    model_config = setup_experiment(experiment)

    # loading dataset
    dataset_module = importlib.import_module(f"ludwig.datasets.{experiment['dataset_name']}")
    dataset = load_from_module(dataset_module, model_config["output_features"][0])

    # running model and capturing metrics
    model = LudwigModel(config=model_config, logging_level=logging.ERROR)
    _, _, _, output_directory = model.experiment(
        dataset=dataset,
        output_directory=experiment["dataset_name"],
        skip_save_processed_input=True,
        skip_save_unprocessed_output=True,
        skip_save_predictions=True,
        skip_collect_predictions=True,
    )
    delete_model_checkpoints(output_directory)
    if export_artifacts_dict["export_artifacts"]:
        export_base_path = export_artifacts_dict["export_base_path"]
        export_artifacts(experiment, output_directory, export_base_path)


def benchmark(bench_config_path: str) -> None:
    """Launch benchmarking suite from a benchmarking config.

    bench_config_path: config for the benchmarking tool. Specifies datasets and their
        corresponding Ludwig configs, as well as export options.
    """
    config = load_yaml(bench_config_path)
    for experiment in config["datasets"]:
        try:
            if "experiment_name" not in experiment:
                experiment["experiment_name"] = config["global_experiment_name"]
            benchmark_one_local(experiment, export_artifacts_dict=config["export"][0])
        except Exception:
            logging.exception(
                f"Experiment *{experiment['experiment_name']}* on dataset *{experiment['dataset_name']}* failed"
            )


def cli(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script runs a ludwig experiment on datasets specified in the benchmark config and exports "
        "the experiment artifact for each of the datasets following the export parameters specified in"
        "the benchmarking config.",
        prog="ludwig benchmark",
        usage="%(prog)s [options]",
    )
    parser.add_argument("--benchmarking_config", type=str, help="The benchmarking config.")
    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)
    benchmark(args.config)



import contextlib
import os
import tempfile
import uuid

import pytest
import yaml

from ludwig.api import LudwigModel
from ludwig.backend import initialize_backend
from ludwig.constants import TRAINER
from ludwig.globals import DESCRIPTION_FILE_NAME
from ludwig.utils import fs_utils
from tests.integration_tests.utils import category_feature, generate_data, private_param, sequence_feature


@contextlib.contextmanager
def remote_tmpdir(fs_protocol, bucket):
    if bucket is None:
        with tempfile.TemporaryDirectory() as tmpdir:
            yield f"{fs_protocol}://{tmpdir}"
        return

    prefix = f"tmp_{uuid.uuid4().hex}"
    tmpdir = f"{fs_protocol}://{bucket}/{prefix}"
    try:
        yield tmpdir
    finally:
        fs_utils.delete(tmpdir, recursive=True)


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("local", id="local"),
        pytest.param("ray", id="ray", marks=pytest.mark.distributed),
    ],
)
@pytest.mark.parametrize(
    "fs_protocol,bucket", [("file", None), private_param(("s3", "ludwig-tests"))], ids=["file", "s3"]
)
def test_remote_training_set(csv_filename, fs_protocol, bucket, backend, ray_cluster_2cpu):
    with remote_tmpdir(fs_protocol, bucket) as tmpdir:
        input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
        output_features = [category_feature(decoder={"vocab_size": 2}, reduce_input="sum")]

        train_csv = os.path.join(tmpdir, "training.csv")
        val_csv = os.path.join(tmpdir, "validation.csv")
        test_csv = os.path.join(tmpdir, "test.csv")

        local_csv = generate_data(input_features, output_features, csv_filename)
        fs_utils.upload_file(local_csv, train_csv)
        fs_utils.copy(train_csv, val_csv)
        fs_utils.copy(train_csv, test_csv)

        config = {
            "input_features": input_features,
            "output_features": output_features,
            "combiner": {"type": "concat", "output_size": 14},
            TRAINER: {"epochs": 2},
        }

        config_path = os.path.join(tmpdir, "config.yaml")
        with fs_utils.open_file(config_path, "w") as f:
            yaml.dump(config, f)

        backend_config = {
            "type": backend,
        }
        backend = initialize_backend(backend_config)

        output_directory = os.path.join(tmpdir, "output")
        model = LudwigModel(config_path, backend=backend)
        _, _, output_run_directory = model.train(
            training_set=train_csv, validation_set=val_csv, test_set=test_csv, output_directory=output_directory
        )

        assert os.path.join(output_directory, "api_experiment_run") == output_run_directory
        assert fs_utils.path_exists(os.path.join(output_run_directory, DESCRIPTION_FILE_NAME))
        assert fs_utils.path_exists(os.path.join(output_run_directory, "training_statistics.json"))
        assert fs_utils.path_exists(os.path.join(output_run_directory, "model"))
        assert fs_utils.path_exists(os.path.join(output_run_directory, "model", "model_weights"))

        model.predict(dataset=test_csv, output_directory=output_directory)

        # Train again, this time the cache will be used
        # Resume from the remote output directory
        model.train(
            training_set=train_csv, validation_set=val_csv, test_set=test_csv, model_resume_path=output_run_directory
        )



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os.path
import shutil
import uuid
from unittest.mock import patch

import pytest
from packaging import version

from ludwig.api import LudwigModel
from ludwig.callbacks import Callback
from ludwig.constants import ACCURACY, AUTO, EXECUTOR, MAX_CONCURRENT_TRIALS, TRAINER
from ludwig.globals import HYPEROPT_STATISTICS_FILE_NAME
from ludwig.hyperopt.results import HyperoptResults
from ludwig.hyperopt.run import hyperopt
from ludwig.hyperopt.utils import update_hyperopt_params_with_defaults
from ludwig.utils.defaults import merge_with_defaults
from tests.integration_tests.utils import binary_feature, create_data_set_to_use, generate_data, number_feature

try:
    import ray

    # Ray nightly version is always set to 3.0.0.dev0
    _ray_nightly = version.parse(ray.__version__) >= version.parse("3.0.0.dev0")
    _ray_114 = version.parse(ray.__version__) >= version.parse("1.14")
    if _ray_114:
        from ray.tune.syncer import get_node_to_storage_syncer, SyncConfig
    else:
        from ray.tune.syncer import get_sync_client

    from ludwig.backend.ray import RayBackend
    from ludwig.hyperopt.execution import _get_relative_checkpoints_dir_parts, RayTuneExecutor
except ImportError:
    ray = None
    _ray_nightly = False
    RayTuneExecutor = object

# Ray mocks

# Dummy sync templates
LOCAL_SYNC_TEMPLATE = "echo {source}/ {target}/"
LOCAL_DELETE_TEMPLATE = "echo {target}"


def mock_storage_client(path):
    """Mocks storage client that treats a local dir as durable storage."""
    os.makedirs(path, exist_ok=True)
    if _ray_114:
        syncer = get_node_to_storage_syncer(SyncConfig(upload_dir=path))
    else:
        syncer = get_sync_client(LOCAL_SYNC_TEMPLATE, LOCAL_DELETE_TEMPLATE)
    return syncer


HYPEROPT_CONFIG = {
    "parameters": {
        "trainer.learning_rate": {
            "space": "loguniform",
            "lower": 0.001,
            "upper": 0.1,
        },
        "combiner.num_fc_layers": {"space": "randint", "lower": 1, "upper": 3},
        "combiner.num_steps": {"space": "grid_search", "values": [1, 2, 3]},
    },
    "goal": "minimize",
}


SCENARIOS = [
    {
        "executor": {"type": "ray", "num_samples": 2, "cpu_resources_per_trial": 1},
        "search_alg": {"type": "variant_generator"},
    },
    # TODO(shreya): Uncomment when https://github.com/ludwig-ai/ludwig/issues/2039 is fixed.
    # {
    #     "type": "ray",
    #     "num_samples": 1,
    #     "scheduler": {
    #         "type": "async_hyperband",
    #         "time_attr": "training_iteration",
    #         "reduction_factor": 2,
    #         "dynamic_resource_allocation": True,
    #     },
    # },
    {
        "executor": {
            "type": "ray",
            "num_samples": 3,
            "scheduler": {
                "type": "hb_bohb",
                "time_attr": "training_iteration",
                "reduction_factor": 4,
            },
            "cpu_resources_per_trial": 1,
        },
        "search_alg": {"type": "bohb"},
    },
]


# TODO ray: replace legacy mode when Ray Train supports placement groups
RAY_BACKEND_KWARGS = {"processor": {"parallelism": 4}}


def _get_config(search_alg, executor):
    input_features = [number_feature(), number_feature()]
    output_features = [binary_feature()]

    return {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "num_fc_layers": 2},
        TRAINER: {"epochs": 1, "learning_rate": 0.001},
        "hyperopt": {
            **HYPEROPT_CONFIG,
            "executor": executor,
            "search_alg": search_alg,
        },
    }


class MockRayTuneExecutor(RayTuneExecutor):
    def _get_sync_client_and_remote_checkpoint_dir(self, trial_dir):
        remote_checkpoint_dir = os.path.join(self.mock_path, *_get_relative_checkpoints_dir_parts(trial_dir))
        return mock_storage_client(remote_checkpoint_dir), remote_checkpoint_dir


class TestCallback(Callback):
    def __init__(self):
        self.preprocessed = False

    def on_hyperopt_preprocessing_start(self, *args, **kwargs):
        self.preprocessed = True

    def on_hyperopt_start(self, *args, **kwargs):
        assert self.preprocessed


@pytest.fixture
def ray_mock_dir():
    path = os.path.join(ray._private.utils.get_user_temp_dir(), f"mock-client-{uuid.uuid4().hex[:4]}") + os.sep
    os.makedirs(path, exist_ok=True)
    try:
        yield path
    finally:
        shutil.rmtree(path)


def run_hyperopt_executor(
    search_alg,
    executor,
    csv_filename,
    ray_mock_dir,
    validate_output_feature=False,
    validation_metric=None,
):
    config = _get_config(search_alg, executor)

    csv_filename = os.path.join(ray_mock_dir, "dataset.csv")
    dataset_csv = generate_data(config["input_features"], config["output_features"], csv_filename, num_examples=100)
    dataset_parquet = create_data_set_to_use("parquet", dataset_csv)

    config = merge_with_defaults(config)

    hyperopt_config = config["hyperopt"]

    if validate_output_feature:
        hyperopt_config["output_feature"] = config["output_features"][0]["name"]
    if validation_metric:
        hyperopt_config["validation_metric"] = validation_metric

    backend = RayBackend(**RAY_BACKEND_KWARGS)
    update_hyperopt_params_with_defaults(hyperopt_config)
    if hyperopt_config[EXECUTOR].get(MAX_CONCURRENT_TRIALS) == AUTO:
        hyperopt_config[EXECUTOR][MAX_CONCURRENT_TRIALS] = backend.max_concurrent_trials(hyperopt_config)

    parameters = hyperopt_config["parameters"]
    if search_alg.get("type", "") == "bohb":
        # bohb does not support grid_search search space
        del parameters["combiner.num_steps"]
        hyperopt_config["parameters"] = parameters

    split = hyperopt_config["split"]
    output_feature = hyperopt_config["output_feature"]
    metric = hyperopt_config["metric"]
    goal = hyperopt_config["goal"]
    search_alg = hyperopt_config["search_alg"]

    # preprocess
    model = LudwigModel(config=config, backend=backend)
    training_set, validation_set, test_set, training_set_metadata = model.preprocess(
        dataset=dataset_parquet,
    )

    # hyperopt
    hyperopt_executor = MockRayTuneExecutor(
        parameters, output_feature, metric, goal, split, search_alg=search_alg, **hyperopt_config[EXECUTOR]
    )
    hyperopt_executor.mock_path = os.path.join(ray_mock_dir, "bucket")

    hyperopt_executor.execute(
        config,
        training_set=training_set,
        validation_set=validation_set,
        test_set=test_set,
        training_set_metadata=training_set_metadata,
        backend=backend,
        output_directory=ray_mock_dir,
        skip_save_processed_input=True,
        skip_save_unprocessed_output=True,
    )


@pytest.mark.skipif(_ray_nightly, reason="https://github.com/ludwig-ai/ludwig/issues/2451")
@pytest.mark.distributed
@pytest.mark.parametrize("scenario", SCENARIOS)
def test_hyperopt_executor(scenario, csv_filename, ray_mock_dir, ray_cluster_7cpu):
    search_alg = scenario["search_alg"]
    executor = scenario["executor"]
    run_hyperopt_executor(search_alg, executor, csv_filename, ray_mock_dir)


@pytest.mark.skip(reason="https://github.com/ludwig-ai/ludwig/issues/1441")
@pytest.mark.distributed
def test_hyperopt_executor_with_metric(csv_filename, ray_mock_dir, ray_cluster_7cpu):
    run_hyperopt_executor(
        # {"type": "ray", "num_samples": 2},
        # {"type": "ray"},
        {"type": "variant_generator"},  # search_alg
        {"type": "ray", "num_samples": 2},  # executor
        csv_filename,
        ray_mock_dir,
        validate_output_feature=True,
        validation_metric=ACCURACY,
    )


@pytest.mark.skip(reason="https://github.com/ludwig-ai/ludwig/issues/1441")
@pytest.mark.distributed
@patch("ludwig.hyperopt.execution.RayTuneExecutor", MockRayTuneExecutor)
def test_hyperopt_run_hyperopt(csv_filename, ray_mock_dir, ray_cluster_7cpu):
    input_features = [number_feature(), number_feature()]
    output_features = [binary_feature()]

    csv_filename = os.path.join(ray_mock_dir, "dataset.csv")
    dataset_csv = generate_data(input_features, output_features, csv_filename, num_examples=100)
    dataset_parquet = create_data_set_to_use("parquet", dataset_csv)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "num_fc_layers": 2},
        TRAINER: {"epochs": 4, "learning_rate": 0.001},
        "backend": {"type": "ray", **RAY_BACKEND_KWARGS},
    }

    output_feature_name = output_features[0]["name"]

    hyperopt_configs = {
        "parameters": {
            "trainer.learning_rate": {
                "space": "loguniform",
                "lower": 0.001,
                "upper": 0.1,
            },
            output_feature_name + ".output_size": {"space": "randint", "lower": 2, "upper": 8},
            output_feature_name + ".num_fc_layers": {"space": "randint", "lower": 1, "upper": 3},
        },
        "goal": "minimize",
        "output_feature": output_feature_name,
        "validation_metrics": "loss",
        "executor": {"type": "ray", "num_samples": 2},
        "search_alg": {"type": "variant_generator"},
    }

    # add hyperopt parameter space to the config
    config["hyperopt"] = hyperopt_configs
    run_hyperopt(config, dataset_parquet, ray_mock_dir)


def run_hyperopt(
    config,
    rel_path,
    out_dir,
    experiment_name="ray_hyperopt",
):
    callback = TestCallback()
    hyperopt_results = hyperopt(
        config,
        dataset=rel_path,
        output_directory=out_dir,
        experiment_name=experiment_name,
        callbacks=[callback],
    )

    # check for return results
    assert isinstance(hyperopt_results, HyperoptResults)

    # check for existence of the hyperopt statistics file
    assert os.path.isfile(os.path.join(out_dir, experiment_name, HYPEROPT_STATISTICS_FILE_NAME))



import os
import random
import string

import numpy as np
import pandas as pd
import pytest
from PIL import Image

from ludwig.api import LudwigModel
from ludwig.constants import COLUMN, DECODER, NAME, PROC_COLUMN, TRAINER
from ludwig.data.concatenate_datasets import concatenate_df
from tests.integration_tests.utils import (
    audio_feature,
    binary_feature,
    category_feature,
    generate_data,
    image_feature,
    LocalTestBackend,
    number_feature,
    sequence_feature,
)

NUM_EXAMPLES = 20


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("local", id="local"),
        pytest.param("ray", id="ray", marks=pytest.mark.distributed),
    ],
)
def test_sample_ratio(backend, tmpdir, ray_cluster_2cpu):
    num_examples = 100
    sample_ratio = 0.25

    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]
    data_csv = generate_data(
        input_features, output_features, os.path.join(tmpdir, "dataset.csv"), num_examples=num_examples
    )
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "trainer": {
            "epochs": 2,
        },
        "preprocessing": {"sample_ratio": sample_ratio},
    }

    model = LudwigModel(config, backend=backend)
    train_set, val_set, test_set, _ = model.preprocess(
        data_csv,
        skip_save_processed_input=True,
    )

    sample_size = num_examples * sample_ratio
    count = len(train_set) + len(val_set) + len(test_set)
    assert sample_size == count


def test_strip_whitespace_category(csv_filename, tmpdir):
    data_csv_path = os.path.join(tmpdir, csv_filename)

    input_features = [binary_feature()]
    cat_feat = category_feature(decoder={"vocab_size": 3})
    output_features = [cat_feat]
    backend = LocalTestBackend()
    config = {"input_features": input_features, "output_features": output_features}

    training_data_csv_path = generate_data(input_features, output_features, data_csv_path)
    df = pd.read_csv(training_data_csv_path)

    # prefix with whitespace
    df[cat_feat[COLUMN]] = df[cat_feat[COLUMN]].apply(lambda s: " " + s)

    # run preprocessing
    ludwig_model = LudwigModel(config, backend=backend)
    train_ds, _, _, metadata = ludwig_model.preprocess(dataset=df)

    # expect values containing whitespaces to be properly mapped to vocab_size unique values
    assert len(np.unique(train_ds.dataset[cat_feat[PROC_COLUMN]])) == cat_feat[DECODER]["vocab_size"]


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("local", id="local"),
        pytest.param("ray", id="ray", marks=pytest.mark.distributed),
    ],
)
def test_with_split(backend, csv_filename, tmpdir, ray_cluster_2cpu):
    num_examples = NUM_EXAMPLES
    train_set_size = int(num_examples * 0.8)
    val_set_size = int(num_examples * 0.1)
    test_set_size = int(num_examples * 0.1)

    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]
    data_csv = generate_data(
        input_features, output_features, os.path.join(tmpdir, csv_filename), num_examples=num_examples
    )
    data_df = pd.read_csv(data_csv)
    data_df["split"] = [0] * train_set_size + [1] * val_set_size + [2] * test_set_size
    data_df.to_csv(data_csv, index=False)
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "trainer": {
            "epochs": 2,
        },
        "preprocessing": {"split": {"type": "fixed"}},
    }

    model = LudwigModel(config, backend=backend)
    train_set, val_set, test_set, _ = model.preprocess(
        data_csv,
        skip_save_processed_input=False,
    )
    assert len(train_set) == train_set_size
    assert len(val_set) == val_set_size
    assert len(test_set) == test_set_size


@pytest.mark.distributed
@pytest.mark.parametrize("feature_fn", [image_feature, audio_feature])
def test_dask_known_divisions(feature_fn, csv_filename, tmpdir, ray_cluster_2cpu):
    import dask.dataframe as dd

    input_features = [feature_fn(os.path.join(tmpdir, "generated_output"))]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]
    data_csv = generate_data(input_features, output_features, os.path.join(tmpdir, csv_filename), num_examples=100)
    data_df = dd.from_pandas(pd.read_csv(data_csv), npartitions=2)
    assert data_df.known_divisions

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "trainer": {
            "epochs": 2,
        },
    }

    backend = "ray"
    model = LudwigModel(config, backend=backend)
    train_set, val_set, test_set, _ = model.preprocess(
        data_df,
        skip_save_processed_input=False,
    )


@pytest.mark.distributed
def test_drop_empty_partitions(csv_filename, tmpdir, ray_cluster_2cpu):
    import dask.dataframe as dd

    input_features = [image_feature(os.path.join(tmpdir, "generated_output"))]
    output_features = [category_feature(vocab_size=5, reduce_input="sum")]

    # num_examples and npartitions set such that each post-split DataFrame has >1 samples, but empty partitions.
    data_csv = generate_data(input_features, output_features, os.path.join(tmpdir, csv_filename), num_examples=25)
    data_df = dd.from_pandas(pd.read_csv(data_csv), npartitions=10)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "trainer": {
            "epochs": 2,
        },
    }

    backend = "ray"
    model = LudwigModel(config, backend=backend)
    train_set, val_set, test_set, _ = model.preprocess(
        data_df,
        skip_save_processed_input=True,
    )
    for dataset in [train_set, val_set, test_set]:
        df = dataset.ds.to_dask()
        for partition in df.partitions:
            assert len(partition) > 0, "empty partitions found in dataset"


@pytest.mark.parametrize("generate_images_as_numpy", [False, True])
def test_read_image_from_path(tmpdir, csv_filename, generate_images_as_numpy):
    input_features = [image_feature(os.path.join(tmpdir, "generated_output"), save_as_numpy=generate_images_as_numpy)]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]
    data_csv = generate_data(
        input_features, output_features, os.path.join(tmpdir, csv_filename), num_examples=NUM_EXAMPLES
    )

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "trainer": {"epochs": 2},
    }

    model = LudwigModel(config)
    model.preprocess(
        data_csv,
        skip_save_processed_input=False,
    )


def test_read_image_from_numpy_array(tmpdir, csv_filename):
    input_features = [image_feature(os.path.join(tmpdir, "generated_output"))]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"epochs": 2},
    }

    data_csv = generate_data(
        input_features, output_features, os.path.join(tmpdir, csv_filename), num_examples=NUM_EXAMPLES
    )

    df = pd.read_csv(data_csv)
    processed_df_rows = []

    for _, row in df.iterrows():
        processed_df_rows.append(
            {
                input_features[0][NAME]: np.array(Image.open(row[input_features[0][NAME]])),
                output_features[0][NAME]: row[output_features[0][NAME]],
            }
        )

    df_with_images_as_numpy_arrays = pd.DataFrame(processed_df_rows)

    model = LudwigModel(config)
    model.preprocess(
        df_with_images_as_numpy_arrays,
        skip_save_processed_input=False,
    )


def test_number_feature_wrong_dtype(csv_filename, tmpdir):
    """Tests that a number feature with all string values is treated as having missing values by default."""
    data_csv_path = os.path.join(tmpdir, csv_filename)

    num_feat = number_feature()
    input_features = [num_feat]
    output_features = [binary_feature()]
    config = {"input_features": input_features, "output_features": output_features}

    training_data_csv_path = generate_data(input_features, output_features, data_csv_path)
    df = pd.read_csv(training_data_csv_path)

    # convert numbers to random strings
    def random_string():
        letters = string.ascii_lowercase
        return "".join(random.choice(letters) for _ in range(10))

    df[num_feat[COLUMN]] = df[num_feat[COLUMN]].apply(lambda _: random_string())

    # run preprocessing
    backend = LocalTestBackend()
    ludwig_model = LudwigModel(config, backend=backend)
    train_ds, val_ds, test_ds, _ = ludwig_model.preprocess(dataset=df)

    concatenated_df = concatenate_df(train_ds.to_df(), val_ds.to_df(), test_ds.to_df(), backend)

    # check that train_ds had invalid values replaced with the missing value
    assert len(concatenated_df) == len(df)
    assert np.all(concatenated_df[num_feat[PROC_COLUMN]] == 0.0)


def test_column_feature_type_mismatch_fill():
    """Tests that we are able to fill missing values even in columns where the column dtype and desired feature
    dtype do not match."""
    cat_feat = category_feature()
    bin_feat = binary_feature()
    input_features = [cat_feat]
    output_features = [bin_feat]
    config = {"input_features": input_features, "output_features": output_features}

    # Construct dataframe with int-like column representing a categorical feature
    df = pd.DataFrame(
        {
            cat_feat[NAME]: pd.Series(pd.array([None] + [1] * 24, dtype=pd.Int64Dtype())),
            bin_feat[NAME]: pd.Series([True] * 25),
        }
    )

    # run preprocessing
    backend = LocalTestBackend()
    ludwig_model = LudwigModel(config, backend=backend)
    train_ds, val_ds, test_ds, _ = ludwig_model.preprocess(dataset=df)


@pytest.mark.parametrize("format", ["file", "df"])
def test_presplit_override(format, tmpdir):
    """Tests that provising a pre-split file or dataframe overrides the user's split config."""
    num_feat = number_feature(normalization=None)
    input_features = [num_feat, sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]

    data_csv = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"), num_examples=25)
    data_df = pd.read_csv(data_csv)

    # Set the feature value equal to an ordinal index so we can ensure the splits are identical before and after
    # preprocessing.
    data_df[num_feat[COLUMN]] = data_df.index

    train_df = data_df[:15]
    val_df = data_df[15:20]
    test_df = data_df[20:]

    train_data = train_df
    val_data = val_df
    test_data = test_df

    if format == "file":
        train_data = os.path.join(tmpdir, "train.csv")
        val_data = os.path.join(tmpdir, "val.csv")
        test_data = os.path.join(tmpdir, "test.csv")

        train_df.to_csv(train_data)
        val_df.to_csv(val_data)
        test_df.to_csv(test_data)

    data_df.to_csv(data_csv, index=False)
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "trainer": {
            "epochs": 2,
        },
        "preprocessing": {"split": {"type": "random"}},
    }

    model = LudwigModel(config, backend=LocalTestBackend())
    train_set, val_set, test_set, _ = model.preprocess(
        training_set=train_data, validation_set=val_data, test_set=test_data
    )

    assert len(train_set) == len(train_df)
    assert len(val_set) == len(val_df)
    assert len(test_set) == len(test_df)

    assert np.all(train_set.to_df()[num_feat[PROC_COLUMN]].values == train_df[num_feat[COLUMN]].values)
    assert np.all(val_set.to_df()[num_feat[PROC_COLUMN]].values == val_df[num_feat[COLUMN]].values)
    assert np.all(test_set.to_df()[num_feat[PROC_COLUMN]].values == test_df[num_feat[COLUMN]].values)


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("local", id="local"),
        pytest.param("ray", id="ray", marks=pytest.mark.distributed),
    ],
)
def test_empty_training_set_error(backend, tmpdir, ray_cluster_2cpu):
    """Tests that an error is raised if one or more of the splits is empty after preprocessing."""
    data_csv_path = os.path.join(tmpdir, "data.csv")

    out_feat = binary_feature()
    input_features = [number_feature()]
    output_features = [out_feat]
    config = {"input_features": input_features, "output_features": output_features}

    training_data_csv_path = generate_data(input_features, output_features, data_csv_path)
    df = pd.read_csv(training_data_csv_path)

    # Convert all the output features rows to null. Because the default missing value strategy is to drop empty output
    # rows, this will result in the dataset being empty after preprocessing.
    df[out_feat[COLUMN]] = None

    ludwig_model = LudwigModel(config, backend=backend)
    with pytest.raises(ValueError, match="Training data is empty following preprocessing"):
        ludwig_model.preprocess(dataset=df)


@pytest.mark.distributed
@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("local", id="local"),
        pytest.param("ray", id="ray", marks=pytest.mark.distributed),
    ],
)
def test_in_memory_dataset_size(backend, tmpdir, ray_cluster_2cpu):
    data_csv_path = os.path.join(tmpdir, "data.csv")

    out_feat = binary_feature()
    input_features = [number_feature()]
    output_features = [out_feat]
    config = {"input_features": input_features, "output_features": output_features}

    training_data_csv_path = generate_data(input_features, output_features, data_csv_path)
    df = pd.read_csv(training_data_csv_path)

    ludwig_model = LudwigModel(config, backend=backend)
    training_dataset, validation_dataset, test_dataset, _ = ludwig_model.preprocess(dataset=df)

    assert training_dataset.in_memory_size_bytes > 0
    assert validation_dataset.in_memory_size_bytes > 0
    assert test_dataset.in_memory_size_bytes > 0



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Callable, Generator, Optional

import torch
import torchmetrics.functional as metrics_F
from torch import Tensor, tensor
from torchmetrics import Accuracy as _Accuracy
from torchmetrics import MeanAbsoluteError
from torchmetrics import MeanMetric as _MeanMetric
from torchmetrics import MeanSquaredError, Metric
from torchmetrics.functional.regression.r2 import _r2_score_compute, _r2_score_update
from torchmetrics.metric import jit_distributed_available

from ludwig.constants import (
    ACCURACY,
    BINARY,
    CATEGORY,
    HITS_AT_K,
    JACCARD,
    LOGITS,
    LOSS,
    MAXIMIZE,
    MEAN_ABSOLUTE_ERROR,
    MEAN_SQUARED_ERROR,
    MINIMIZE,
    NUMBER,
    PREDICTIONS,
    PROBABILITIES,
    R2,
    ROC_AUC,
    ROOT_MEAN_SQUARED_ERROR,
    ROOT_MEAN_SQUARED_PERCENTAGE_ERROR,
    SEQUENCE,
    SET,
    TEXT,
    TOKEN_ACCURACY,
    VECTOR,
)
from ludwig.modules.loss_modules import (
    BWCEWLoss,
    SequenceSoftmaxCrossEntropyLoss,
    SigmoidCrossEntropyLoss,
    SoftmaxCrossEntropyLoss,
)
from ludwig.modules.metric_registry import metric_registry, register_metric
from ludwig.utils.horovod_utils import gather_all_tensors, is_distributed_available
from ludwig.utils.loss_utils import rmspe_loss
from ludwig.utils.metric_utils import masked_correct_predictions
from ludwig.utils.torch_utils import sequence_length_2D

logger = logging.getLogger(__name__)


class LudwigMetric(Metric, ABC):
    @classmethod
    def can_report(cls, feature: "OutputFeature") -> bool:  # noqa: F821
        return True

    @classmethod
    @abstractmethod
    def get_objective(cls):
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def get_inputs(cls):
        """Returns the key of the tensor from the predictions() Dict that should be used for computing this metric.

        For example: PREDICTIONS would be used for accuracy metrics while LOGITS would be used for loss metrics.
        """
        raise NotImplementedError()

    @contextmanager
    def sync_context(
        self,
        dist_sync_fn: Optional[Callable] = None,
        process_group: Optional[Any] = None,
        should_sync: bool = True,
        should_unsync: bool = True,
        distributed_available: Optional[Callable] = jit_distributed_available,
    ) -> Generator:
        """Override the behavior of this in the base class to support Horovod."""
        self.sync(
            dist_sync_fn=gather_all_tensors,
            process_group=process_group,
            should_sync=should_sync,
            distributed_available=is_distributed_available,
        )

        yield

        self.unsync(should_unsync=self._is_synced and should_unsync)


@register_metric(ROOT_MEAN_SQUARED_ERROR, [NUMBER])
class RMSEMetric(MeanSquaredError, LudwigMetric):
    """Root mean squared error metric."""

    def __init__(self, **kwargs):
        super().__init__(squared=False, dist_sync_fn=gather_all_tensors, **kwargs)

    @classmethod
    def get_objective(cls):
        return MINIMIZE

    @classmethod
    def get_inputs(cls):
        return PREDICTIONS


@register_metric(ROC_AUC, [BINARY])
class ROCAUCMetric(LudwigMetric):
    """Fast implementation of metric for area under ROC curve."""

    def __init__(
        self,
        num_thresholds: int = 201,
        epsilon: float = 1e-7,
        **kwargs,
    ) -> None:
        super().__init__(dist_sync_fn=gather_all_tensors)
        self.num_thresholds = num_thresholds
        self.epsilon = epsilon
        self.add_state("summary_stats", torch.zeros(num_thresholds, 4), dist_reduce_fx="sum")

    def _get_thresholds(self, device, dtype) -> Tensor:
        thresholds = torch.linspace(0, 1, self.num_thresholds, device=device, dtype=dtype)
        thresholds[0] -= self.epsilon
        thresholds[-1] += self.epsilon
        return thresholds

    def update(self, preds: Tensor, target: Tensor) -> None:
        # Currently only supported for binary tasks.
        if preds.ndim > 1 or target.ndim > 1:
            raise RuntimeError(
                f"Only binary tasks supported, but received input of "
                f"{max(preds.ndim, target.ndim)} dimensions while expecting"
                f"1-dimensional input."
            )

        if torch.min(preds) < 0 or torch.max(preds) > 1:
            raise RuntimeError(
                f"Only binary tasks supported, but received predictions in range "
                f"({torch.min(preds)}, {torch.max(preds)})."
            )

        thresholds = self._get_thresholds(preds.device, preds.dtype)
        target = target.to(bool).type(preds.dtype)

        preds = preds.unsqueeze(1)
        target = target.unsqueeze(1)

        # Compute correct predictions at each threshold.
        correct_predictions = ((preds >= thresholds) == target).to(int)

        # Compute true positives, false positives, true negatives, false negatives.
        # overall_predictions is a tensor where each cell represents the type of prediction:
        # 0: false positive
        # 1: true negative
        # 2: false negative
        # 3: true positive
        overall_predictions = correct_predictions + (2 * target)

        # Sum up the number of true positives, false positives, true negatives, false negatives at each threshold.
        self.summary_stats += torch.eye(4, device=preds.device)[overall_predictions.T.long()].sum(dim=1, keepdim=False)

    def compute(self) -> Tensor:
        # Compute true positives, false positives, true negatives, false negatives.
        self.summary_stats = self.summary_stats.squeeze()
        false_positives = self.summary_stats[:, 0]
        true_negatives = self.summary_stats[:, 1]
        false_negatives = self.summary_stats[:, 2]
        true_positives = self.summary_stats[:, 3]

        true_positive_rate = true_positives / (true_positives + false_negatives)
        false_positive_rate = false_positives / (false_positives + true_negatives)

        # Compute area under ROC curve. Multiply by -1 because tpr and fpr are computed from the opposite direction.
        return -1 * torch.trapz(true_positive_rate, false_positive_rate)

    @classmethod
    def get_objective(cls):
        return MAXIMIZE

    @classmethod
    def get_inputs(cls):
        return PROBABILITIES


class MeanMetric(LudwigMetric):
    """Abstract class for computing mean of metrics."""

    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=gather_all_tensors)
        self.avg = _MeanMetric()

    def update(self, preds: Tensor, target: Tensor) -> None:
        self.avg.update(self.get_current_value(preds, target))

    def compute(self) -> Tensor:
        return self.avg.compute()

    def reset(self):
        super().reset()
        self.avg.reset()

    @abstractmethod
    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        raise NotImplementedError()


@register_metric(ROOT_MEAN_SQUARED_PERCENTAGE_ERROR, [NUMBER])
class RMSPEMetric(MeanMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=gather_all_tensors)

    """ Root mean squared percentage error metric. """

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        return rmspe_loss(target, preds)

    @classmethod
    def get_objective(cls):
        return MINIMIZE

    @classmethod
    def get_inputs(cls):
        return PREDICTIONS


@register_metric(R2, [NUMBER, VECTOR])
class R2Score(LudwigMetric):
    """Custom R-squared metric implementation that modifies torchmetrics R-squared implementation to return Nan
    when there is only sample. This is because R-squared is only defined for two or more samples.

    Custom implementation uses code from torchmetrics v0.9.2's implementation of R2:
    https://github.com/Lightning-AI/metrics/blob/master/src/torchmetrics/regression/r2.py
    """

    def __init__(
        self, num_outputs: int = 1, adjusted: int = 0, multioutput: str = "uniform_average", **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)

        self.num_outputs = num_outputs

        if adjusted < 0 or not isinstance(adjusted, int):
            raise ValueError("`adjusted` parameter should be an integer larger or equal to 0.")
        self.adjusted = adjusted

        allowed_multioutput = ("raw_values", "uniform_average", "variance_weighted")
        if multioutput not in allowed_multioutput:
            raise ValueError(
                f"Invalid input to argument `multioutput`. Choose one of the following: {allowed_multioutput}"
            )
        self.multioutput = multioutput

        self.add_state("sum_squared_error", default=torch.zeros(self.num_outputs), dist_reduce_fx="sum")
        self.add_state("sum_error", default=torch.zeros(self.num_outputs), dist_reduce_fx="sum")
        self.add_state("residual", default=torch.zeros(self.num_outputs), dist_reduce_fx="sum")
        self.add_state("total", default=tensor(0), dist_reduce_fx="sum")

    def update(self, preds: Tensor, target: Tensor) -> None:
        """Update state with predictions and targets.

        Args:
            preds: Predictions from model
            target: Ground truth values
        """
        sum_squared_error, sum_error, residual, n_obs = _r2_score_update(preds, target)

        self.sum_squared_error += sum_squared_error
        self.sum_error += sum_error
        self.residual += residual
        self.total += n_obs

    def compute(self) -> Tensor:
        """Computes r2 score over the metric states."""

        # self.total maps to the number of observations in preds/target computed during update()
        if self.total <= 1:
            logger.warning(
                """R-squared (r2) is not defined for one sample. It needs at least two samples. Returning NaN."""
            )
            return torch.tensor(float("nan"))

        return _r2_score_compute(
            self.sum_squared_error, self.sum_error, self.residual, self.total, self.adjusted, self.multioutput
        )

    @classmethod
    def get_objective(cls):
        return MAXIMIZE

    @classmethod
    def get_inputs(cls):
        return PREDICTIONS


@register_metric(LOSS, [])
class LossMetric(MeanMetric, ABC):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        raise NotImplementedError()

    @classmethod
    def get_objective(cls):
        return MINIMIZE

    @classmethod
    def get_inputs(cls):
        return LOGITS

    @classmethod
    def can_report(cls, feature: "OutputFeature") -> bool:  # noqa: F821
        return False


@register_metric("binary_weighted_cross_entropy", [BINARY])
class BWCEWLMetric(LossMetric):
    """Binary Weighted Cross Entropy Weighted Logits Score Metric."""

    def __init__(
        self,
        positive_class_weight: Optional[Tensor] = None,
        robust_lambda: int = 0,
        confidence_penalty: int = 0,
        **kwargs,
    ):
        super().__init__()

        self.loss_function = BWCEWLoss(
            positive_class_weight=positive_class_weight,
            robust_lambda=robust_lambda,
            confidence_penalty=confidence_penalty,
        )

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        return self.loss_function(preds, target)


@register_metric("softmax_cross_entropy", [CATEGORY])
class SoftmaxCrossEntropyMetric(LossMetric):
    def __init__(self, **kwargs):
        super().__init__()
        self.softmax_cross_entropy_function = SoftmaxCrossEntropyLoss(**kwargs)

    def get_current_value(self, preds: Tensor, target: Tensor):
        return self.softmax_cross_entropy_function(preds, target)


@register_metric("sequence_softmax_cross_entropy", [SEQUENCE, TEXT])
class SequenceSoftmaxCrossEntropyMetric(LossMetric):
    def __init__(self, **kwargs):
        super().__init__()
        self.sequence_softmax_cross_entropy_function = SequenceSoftmaxCrossEntropyLoss(**kwargs)

    def get_current_value(self, preds: Tensor, target: Tensor):
        return self.sequence_softmax_cross_entropy_function(preds, target)


@register_metric("sigmoid_cross_entropy", [SET])
class SigmoidCrossEntropyMetric(LossMetric):
    def __init__(self, **kwargs):
        super().__init__()
        self.sigmoid_cross_entropy_function = SigmoidCrossEntropyLoss(**kwargs)

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        return self.sigmoid_cross_entropy_function(preds, target)


@register_metric(TOKEN_ACCURACY, [SEQUENCE, TEXT])
class TokenAccuracyMetric(MeanMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=gather_all_tensors)

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        target = target.type(preds.dtype)
        target_sequence_length = sequence_length_2D(target)
        masked_correct_preds = masked_correct_predictions(target, preds, target_sequence_length)
        return torch.mean(masked_correct_preds)

    @classmethod
    def get_objective(cls):
        return MAXIMIZE

    @classmethod
    def get_inputs(cls):
        return PREDICTIONS


@register_metric(ACCURACY, [BINARY])
class Accuracy(_Accuracy, LudwigMetric):
    """R-squared metric."""

    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=gather_all_tensors)

    @classmethod
    def get_objective(cls):
        return MAXIMIZE

    @classmethod
    def get_inputs(cls):
        return PREDICTIONS


@register_metric(ACCURACY, [CATEGORY])
class CategoryAccuracy(_Accuracy, LudwigMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=gather_all_tensors)

    def update(self, preds: Tensor, target: Tensor) -> None:
        # make sure y_true is tf.int64
        super().update(preds, target.type(torch.long))

    @classmethod
    def get_objective(cls):
        return MAXIMIZE

    @classmethod
    def get_inputs(cls):
        # TODO: double check
        return PREDICTIONS


@register_metric(HITS_AT_K, [CATEGORY])
class HitsAtKMetric(_Accuracy, LudwigMetric):
    def __init__(self, top_k: int = 3, **kwargs):
        super().__init__(top_k=top_k, dist_sync_fn=gather_all_tensors)

    def update(self, preds: Tensor, target: Tensor) -> None:
        super().update(preds, target)

    @classmethod
    def get_objective(cls):
        return MAXIMIZE

    @classmethod
    def get_inputs(cls):
        return LOGITS

    @classmethod
    def can_report(cls, feature: "OutputFeature") -> bool:  # noqa: F821
        return feature.num_classes > feature.top_k


@register_metric(MEAN_ABSOLUTE_ERROR, [NUMBER, VECTOR])
class MAEMetric(MeanAbsoluteError, LudwigMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=gather_all_tensors)

    def update(self, preds: Tensor, target: Tensor) -> None:
        super().update(preds.detach(), target)

    @classmethod
    def get_objective(cls):
        return MINIMIZE

    @classmethod
    def get_inputs(cls):
        return PREDICTIONS


@register_metric(MEAN_SQUARED_ERROR, [NUMBER, VECTOR])
class MSEMetric(MeanSquaredError, LudwigMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=gather_all_tensors)

    def update(self, preds: Tensor, target: Tensor) -> None:
        super().update(preds, target)

    @classmethod
    def get_objective(cls):
        return MINIMIZE

    @classmethod
    def get_inputs(cls):
        return PREDICTIONS


@register_metric(JACCARD, [SET])
class JaccardMetric(MeanMetric):
    def __init__(self, threshold: float = 0.5, **kwargs):
        super().__init__(dist_sync_fn=gather_all_tensors)
        self.threshold = threshold
        self.add_state(name="loss", default=[], dist_reduce_fx="mean")

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        # notation: b is batch size and nc is number of unique elements in the set
        # preds: shape [b, nc] probabilities for each class
        # target: shape [b, nc] bit-mapped set representation
        preds = torch.greater_equal(preds, self.threshold)  # now bit-mapped set
        target = target.type(torch.bool)

        intersection = torch.sum(torch.logical_and(target, preds).type(torch.float32), dim=-1)
        union = torch.sum(torch.logical_or(target, preds).type(torch.float32), dim=-1)

        return intersection / union  # shape [b]

    @classmethod
    def get_objective(cls):
        return MAXIMIZE

    @classmethod
    def get_inputs(cls):
        return PROBABILITIES


def get_improved_fun(metric: str) -> Callable:
    if metric_registry[metric].get_objective() == MINIMIZE:
        return lambda x, y: x < y
    else:
        return lambda x, y: x > y


def get_initial_validation_value(metric: str) -> float:
    if metric_registry[metric].get_objective() == MINIMIZE:
        return float("inf")
    else:
        return float("-inf")


def get_best_function(metric: str) -> Callable:
    if metric_registry[metric].get_objective() == MINIMIZE:
        return min
    else:
        return max


def accuracy(preds: Tensor, target: Tensor) -> Tensor:
    """
    Returns:
        Accuracy (float tensor of shape (1,)).
    """
    return metrics_F.accuracy(preds, target)
    # correct_predictions = predictions == target
    # accuracy = torch.mean(correct_predictions.type(torch.float32))
    # return accuracy, correct_predictions


def perplexity(cross_entropy_loss):
    # This seem weird but is correct:
    # we are returning the cross entropy loss as it will be later summed,
    # divided by the size of the dataset and finally exponentiated,
    # because perplexity has a avg_exp aggregation strategy
    # in the output config in SequenceOutputFeature.
    # This implies that in Model update_output_stats_batch()
    # the values read from the perplexity node will be summed
    # and in Model update_output_stats() they will be divided
    # by the set size first and exponentiated.
    return cross_entropy_loss


def error(preds: Tensor, target: Tensor) -> Tensor:
    return target - preds


def absolute_error(preds: Tensor, target: Tensor) -> Tensor:
    return torch.abs(target - preds)


def squared_error(preds: Tensor, target: Tensor) -> Tensor:
    return (target - preds) ** 2


def r2(preds: Tensor, target: Tensor) -> Tensor:
    return metrics_F.r2_score(preds, target)



#! /usr/bin/env python
# Copyright (c) 2021 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from ludwig.datasets.base_dataset import BaseDataset, DEFAULT_CACHE_LOCATION
from ludwig.datasets.mixins.kaggle import KaggleDownloadMixin
from ludwig.datasets.mixins.load import CSVLoadMixin
from ludwig.datasets.mixins.process import IdentityProcessMixin
from ludwig.datasets.registry import register_dataset


def load(cache_dir=DEFAULT_CACHE_LOCATION, split=False, kaggle_username=None, kaggle_key=None):
    dataset = SantanderCustomerSatisfaction(cache_dir=cache_dir, kaggle_username=kaggle_username, kaggle_key=kaggle_key)
    return dataset.load(split=split)


@register_dataset(name="santander_customer_satisfaction")
class SantanderCustomerSatisfaction(CSVLoadMixin, IdentityProcessMixin, KaggleDownloadMixin, BaseDataset):
    """Santander Customer Satisfaction Prediction.

    https://www.kaggle.com/c/santander-customer-satisfaction/overview
    """

    def __init__(self, cache_dir=DEFAULT_CACHE_LOCATION, kaggle_username=None, kaggle_key=None):
        self.kaggle_username = kaggle_username
        self.kaggle_key = kaggle_key
        self.is_kaggle_competition = True
        super().__init__(dataset_name="santander_customer_satisfaction", cache_dir=cache_dir)



from typing import Any, Dict, List, Optional, Union

from marshmallow_dataclass import dataclass

from ludwig.schema import utils as schema_utils


@dataclass
class CommonTransformerConfig:
    """Common transformer parameter values."""

    num_layers: int = schema_utils.PositiveInteger(default=1, description="")

    hidden_size: int = schema_utils.NonNegativeInteger(
        default=256,
        description=(
            "The number of hidden units of the TransformerStack as well as the dimension that each incoming input "
            "feature is projected to before feeding to the TransformerStack"
        ),
    )

    num_heads: int = schema_utils.NonNegativeInteger(
        default=8, description="Number of heads of the self attention in the transformer block."
    )

    transformer_output_size: int = schema_utils.NonNegativeInteger(
        default=256,
        description=(
            "Size of the fully connected layer after self attention in the transformer block. This is usually the same "
            "as `hidden_size` and `embedding_size`."
        ),
    )

    dropout: float = schema_utils.FloatRange(
        default=0.1, min=0, max=1, description="Dropout rate for the transformer block."
    )

    fc_layers: Optional[List[Dict[str, Any]]] = schema_utils.DictList(description="")

    # TODO(#1673): Add conditional logic for fields like this one:
    num_fc_layers: int = schema_utils.NonNegativeInteger(
        default=0,
        description="The number of stacked fully connected layers (only applies if `reduce_output` is not null).",
    )

    output_size: int = schema_utils.PositiveInteger(default=256, description="Output size of a fully connected layer.")

    use_bias: bool = schema_utils.Boolean(default=True, description="Whether the layer uses a bias vector.")

    weights_initializer: Union[str, Dict] = schema_utils.InitializerOrDict(default="xavier_uniform", description="")

    bias_initializer: Union[str, Dict] = schema_utils.InitializerOrDict(default="zeros", description="")

    norm: Optional[str] = schema_utils.StringOptions(["batch", "layer"], description="")

    norm_params: Optional[dict] = schema_utils.Dict(description="")

    fc_activation: str = schema_utils.ActivationOptions(default="relu", description="")

    fc_dropout: float = schema_utils.FloatRange(default=0.0, min=0, max=1, description="")

    fc_residual: bool = schema_utils.Boolean(default=False, description="")
